"""Generate code to perform a comparison of deep equality on instances."""

import io
from typing import List, Optional

from icontract import ensure

from aas_core_codegen import intermediate
from aas_core_codegen.common import (
    Stripped,
    Identifier,
    indent_but_first_line,
    assert_never,
)
from aas_core_codegen.golang import (
    common as golang_common,
    naming as golang_naming,
    pointering as golang_pointering,
)
from aas_core_codegen.golang.common import (
    INDENT as I,
    INDENT2 as II,
)


def _generate_for_cls(cls: intermediate.ConcreteClass) -> Stripped:
    """Generate code for the deep equality function for ``cls``."""
    blocks = []  # type: List[Stripped]

    for prop in cls.properties:
        that_var = golang_naming.variable_name(Identifier(f"that_{prop.name}"))
        other_var = golang_naming.variable_name(Identifier(f"other_{prop.name}"))

        getter_name = golang_naming.getter_name(prop.name)

        subblocks = [
            Stripped(
                f"""\
{that_var} := that.{getter_name}()
{other_var} := other.{getter_name}()"""
            )
        ]  # type: List[Stripped]

        if isinstance(prop.type_annotation, intermediate.OptionalTypeAnnotation):
            subblocks.append(
                Stripped(
                    f"""\
if
{I}({that_var} == nil && {other_var} != nil) ||
{I}({that_var} != nil && {other_var} == nil) {{
{I}return false
}}"""
                )
            )

        cmp_subblock = None  # type: Optional[Stripped]

        type_anno = intermediate.beneath_optional(prop.type_annotation)

        if isinstance(type_anno, intermediate.PrimitiveTypeAnnotation) or (
            isinstance(type_anno, intermediate.OurTypeAnnotation)
            and isinstance(
                type_anno.our_type,
                (intermediate.ConstrainedPrimitive, intermediate.Enumeration),
            )
        ):
            primitive_type = intermediate.try_primitive_type(type_anno)

            if golang_pointering.is_pointer_type(prop.type_annotation):
                cmp_subblock = Stripped(
                    f"""\
if *{that_var} != *{other_var} {{
{I}return false
}}"""
                )
            else:
                if primitive_type is intermediate.PrimitiveType.BYTEARRAY:
                    cmp_subblock = Stripped(
                        f"""\
if !bytes.Equal(
{I}{that_var},
{I}{other_var},
) {{
{I}return false
}}"""
                    )
                else:
                    cmp_subblock = Stripped(
                        f"""\
if {that_var} != {other_var} {{
{I}return false
}}"""
                    )

        elif isinstance(type_anno, intermediate.OurTypeAnnotation):
            if isinstance(type_anno.our_type, intermediate.Enumeration):
                raise AssertionError("Should have been handled before")

            elif isinstance(type_anno.our_type, intermediate.ConstrainedPrimitive):
                raise AssertionError("Should have been handled before")

            elif isinstance(
                type_anno.our_type,
                (intermediate.AbstractClass, intermediate.ConcreteClass),
            ):
                cmp_subblock = Stripped(
                    f"""\
if !DeepEqual(
{I}{that_var},
{I}{other_var},
) {{
{I}return false
}}"""
                )

            else:
                # noinspection PyTypeChecker
                assert_never(type_anno.our_type)
        elif isinstance(type_anno, intermediate.ListTypeAnnotation):
            assert isinstance(
                type_anno.items, intermediate.OurTypeAnnotation
            ) and isinstance(
                type_anno.items.our_type,
                (intermediate.AbstractClass, intermediate.ConcreteClass),
            ), (
                f"NOTE (mristin): We expect only lists of classes "
                f"at the moment, but you specified {type_anno}. "
                f"Please contact the developers if you need this feature."
            )

            cmp_subblock = Stripped(
                f"""\
if 
{I}len({that_var}) !=
{I}len({other_var}) {{
{I}return false
}}
for i := range {that_var} {{
{I}if !DeepEqual(
{II}{that_var}[i],
{II}{other_var}[i],
{I}) {{
{II}return false
{I}}}
}}"""
            )
        else:
            # noinspection PyTypeChecker
            assert_never(type_anno)

        assert cmp_subblock is not None

        if isinstance(prop.type_annotation, intermediate.OptionalTypeAnnotation):
            cmp_subblock = Stripped(
                f"""\
if {that_var} != nil {{
{I}{indent_but_first_line(cmp_subblock, I)}
}}"""
            )

        subblocks.append(cmp_subblock)

        blocks.append(Stripped("\n".join(subblocks)))

    blocks.append(Stripped("return true"))

    body = "\n\n".join(blocks)

    interface_name = golang_naming.interface_name(cls.name)

    function_name = golang_naming.private_function_name(
        Identifier(f"deep_equal_{cls.name}")
    )

    return Stripped(
        f"""\
// Perform a comparison for deep equality between `that` and `other` instance.
//
// The deep equality means that all the properties are checked for equality recursively.
func {function_name}(
{I}that aastypes.{interface_name},
{I}other aastypes.{interface_name},
) bool {{
{I}{indent_but_first_line(body, I)}
}}"""
    )


def _generate_dispatch_function(symbol_table: intermediate.SymbolTable) -> Stripped:
    """Generate the main ``DeepEqual`` function which then dispatches as necessary."""
    blocks = [
        Stripped(
            f"""\
if that.ModelType() != other.ModelType() {{
{I}return false
}}"""
        )
    ]

    cases = []  # type: List[Stripped]
    for cls in symbol_table.concrete_classes:
        model_type_literal = golang_naming.enum_literal_name(
            enumeration_name=Identifier("Model_type"), literal_name=cls.name
        )

        deep_equal_func_name = golang_naming.private_function_name(
            Identifier(f"deep_equal_{cls.name}")
        )

        interface_name = golang_naming.interface_name(cls.name)

        cases.append(
            Stripped(
                f"""\
case aastypes.{model_type_literal}:
{I}return {deep_equal_func_name}(
{II}that.(aastypes.{interface_name}),
{II}other.(aastypes.{interface_name}),
{I})"""
            )
        )

    cases_joined = "\n".join(cases)

    switch_stmt = Stripped(
        f"""\
switch that.ModelType() {{
{cases_joined}
}}"""
    )

    blocks.append(switch_stmt)

    blocks.append(
        Stripped(
            """\
panic(fmt.Sprintf("Unexpected model type: %d", that.ModelType()))"""
        )
    )

    body = "\n\n".join(blocks)

    return Stripped(
        f"""\
func DeepEqual(
{I}that aastypes.IClass,
{I}other aastypes.IClass,
) bool {{
{I}{indent_but_first_line(body, I)}
}}"""
    )


# fmt: off
@ensure(
    lambda result: result.endswith('\n'),
    "Trailing newline mandatory for valid end-of-files"
)
# fmt: on
def generate(symbol_table: intermediate.SymbolTable, repo_url: Stripped) -> str:
    """Generate code to perform a comparison of deep equality on instances."""
    blocks = [
        Stripped("package aastesting"),
        golang_common.WARNING,
        Stripped(
            f"""\
import (
{I}"bytes"
{I}"fmt"
{I}aastypes "{repo_url}/types"
)"""
        ),
    ]  # type: List[Stripped]

    for cls in symbol_table.concrete_classes:
        blocks.append(_generate_for_cls(cls=cls))

    blocks.append(_generate_dispatch_function(symbol_table=symbol_table))

    blocks.append(golang_common.WARNING)

    writer = io.StringIO()
    for i, block in enumerate(blocks):
        if i > 0:
            writer.write("\n\n")

        writer.write(block)

    writer.write("\n")

    return writer.getvalue()


assert generate.__doc__ is not None
assert generate.__doc__.strip().startswith(__doc__.strip())



"""Generate code to test copying."""

import io
import textwrap
from typing import List, Optional

from aas_core_codegen import intermediate
from aas_core_codegen.common import (
    Stripped,
    indent_but_first_line,
    assert_never,
    Identifier,
)
from aas_core_codegen.java import common as java_common, naming as java_naming
from aas_core_codegen.java.common import INDENT as I, INDENT2 as II, INDENT3 as III


def _generate_shallow_equals(cls: intermediate.ConcreteClass) -> Stripped:
    """Generate the code for a static shallow ``Equals`` method."""
    if cls.is_implementation_specific:
        raise AssertionError(
            f"(empwilli):"
            f"the class {cls.name!r} is implementation specific. "
            f"at the moment, we assume that all classes are not "
            f"implementation-specific, so that we can automatically generate the "
            f"shallow-equals methods. this way we can dispense of the whole "
            f"snippet/specific-implementation loading logic in "
            f"the unit test generation. please notify the developers if you see this, "
            f"so that we can add the logic for implementation-specific classes "
            f"to this generation script."
        )

    exprs = []  # type: List[str]
    for prop in cls.properties:
        prop_name = java_naming.method_name(Identifier(f"get_{prop.name}"))
        exprs.append(f"that.{prop_name}().equals(other.{prop_name}())")

    # NOTE (empwilli):
    # This is a poor man's line re-flowing.
    exprs_joined = " && ".join(exprs)
    if len(exprs_joined) < 70:
        statement = Stripped(f"return {exprs_joined};")
    else:
        exprs_joined = "\n&& ".join(exprs)
        statement = Stripped(
            f"""\
return (
{I}{indent_but_first_line(exprs_joined, I)});"""
        )

    cls_name_java = java_naming.class_name(cls.name)

    return Stripped(
        f"""\
private static Boolean {cls_name_java}ShallowEquals(
{I}{cls_name_java} that,
{I}{cls_name_java} other) {{
{I}{indent_but_first_line(statement, I)}
}}"""
    )


def _generate_transform_as_deep_equals(cls: intermediate.ConcreteClass) -> Stripped:
    """Generate the transform method that checks for deep equality."""
    if cls.is_implementation_specific:
        raise AssertionError(
            f"(empwilli): "
            f"The class {cls.name!r} is implementation specific. "
            f"At the moment, we assume that all classes are not "
            f"implementation-specific, so that we can automatically generate the "
            f"shallow-equals methods. This way we can dispense of the whole "
            f"snippet/specific-implementation loading logic in "
            f"the unit test generation. Please notify the developers if you see this, "
            f"so that we can add the logic for implementation-specific classes "
            f"to this generation script."
        )

    cls_name = java_naming.class_name(cls.name)

    exprs = []  # type: List[Stripped]

    for prop in cls.properties:
        optional = isinstance(prop.type_annotation, intermediate.OptionalTypeAnnotation)
        type_anno = intermediate.beneath_optional(prop.type_annotation)
        getter_name = java_naming.getter_name(prop.name)
        primitive_type = intermediate.try_primitive_type(type_anno)

        expr = None  # type: Optional[Stripped]

        if isinstance(type_anno, intermediate.PrimitiveTypeAnnotation) or (
            isinstance(type_anno, intermediate.OurTypeAnnotation)
            and isinstance(type_anno.our_type, intermediate.ConstrainedPrimitive)
        ):
            assert primitive_type is not None
            if (
                primitive_type is intermediate.PrimitiveType.BOOL
                or primitive_type is intermediate.PrimitiveType.INT
                or primitive_type is intermediate.PrimitiveType.FLOAT
                or primitive_type is intermediate.PrimitiveType.STR
            ):
                expr = Stripped(
                    f"""\
that.{getter_name}().equals(casted.{getter_name}())"""
                )
            elif primitive_type is intermediate.PrimitiveType.BYTEARRAY:
                expr = Stripped(
                    f"""\
Arrays.equals(
{I}that.{getter_name}().get(),
{I}casted.{getter_name}().get())"""
                )
            else:
                assert_never(primitive_type)

        elif isinstance(type_anno, intermediate.OurTypeAnnotation):
            if isinstance(type_anno.our_type, intermediate.Enumeration):
                expr = Stripped(
                    f"""\
that.{getter_name}().equals(casted.{getter_name}())"""
                )
            elif isinstance(type_anno.our_type, intermediate.ConstrainedPrimitive):
                raise AssertionError("Expected to handle this case above")
            elif isinstance(
                type_anno.our_type,
                (intermediate.AbstractClass, intermediate.ConcreteClass),
            ):
                if optional:
                    expr = Stripped(
                        f"""\
(that.{getter_name}().isPresent()
{I}? casted.{getter_name}().isPresent()
{I}&& transform( that.{getter_name}().get(), casted.{getter_name}().get())
{I}: ! casted.{getter_name}().isPresent())"""
                    )
                else:
                    expr = Stripped(
                        f"""\
transform(
{I}that.{getter_name}(),
{I}casted.{getter_name}())"""
                    )
        elif isinstance(type_anno, intermediate.ListTypeAnnotation):
            assert isinstance(
                type_anno.items, intermediate.OurTypeAnnotation
            ) and isinstance(type_anno.items.our_type, intermediate.Class), (
                f"(empwilli): We handle only lists of classes in the deep "
                f"equality checks at the moment. The meta-model does not contain "
                f"any other lists, so we wanted to keep the code as simple as "
                f"possible, and avoid unrolling. Please contact the developers "
                f"if you need this feature. The class in question was {cls.name!r} and "
                f"the property {prop.name!r}."
            )
            expr = Stripped(
                f"""\
that.{getter_name}().equals(casted.{getter_name}())"""
            )
        else:
            # noinspection PyTypeChecker
            assert_never(type_anno)

        exprs.append(expr)

    body_writer = io.StringIO()
    body_writer.write("return (")
    for i, expr in enumerate(exprs):
        body_writer.write("\n")
        if i > 0:
            body_writer.write(f"{I}&& {indent_but_first_line(expr, I)}")
        else:
            body_writer.write(f"{I}{indent_but_first_line(expr, I)}")

    body_writer.write(");")

    interface_name = java_naming.interface_name(cls.name)
    transform_name = java_naming.method_name(Identifier(f"transform_{cls.name}"))

    return Stripped(
        f"""\
@Override
public Boolean {transform_name}({interface_name} that, IClass other) {{
{I}if (!(other instanceof {cls_name})) {{
{II}return false;
{I}}}

{I}{cls_name} casted = ({cls_name}) that;

{I}{indent_but_first_line(body_writer.getvalue(), I)}
}}"""
    )


def _generate_deep_equals_transformer(
    symbol_table: intermediate.SymbolTable,
) -> Stripped:
    """Generate the transformer that checks for deep equality."""
    blocks = []  # type: List[Stripped]

    for concrete_cls in symbol_table.concrete_classes:
        if concrete_cls.is_implementation_specific:
            raise AssertionError(
                f"(empwilli): "
                f"The class {concrete_cls.name!r} is implementation specific. "
                f"At the moment, we assume that all classes are not "
                f"implementation-specific, so that we can automatically generate the "
                f"deep-equals methods. This way we can dispense of the whole "
                f"snippet/specific-implementation loading logic in "
                f"the unit test generation. Please notify the developers if you see "
                f"this, so that we can add the logic for implementation-specific "
                f"classes to this generation script."
            )

        blocks.append(_generate_transform_as_deep_equals(cls=concrete_cls))

    writer = io.StringIO()
    writer.write(
        """\
private static class DeepEqualiser extends AbstractTransformerWithContext<IClass, Boolean> {
"""
    )

    for i, block in enumerate(blocks):
        if i > 0:
            writer.write("\n\n")

        writer.write(textwrap.indent(block, I))

    writer.write("\n} // class DeepEqualiser")

    return Stripped(writer.getvalue())


def _generate_deep_equals(cls: intermediate.ConcreteClass) -> Stripped:
    """Generate the code for a static deep ``Equals`` method."""
    if cls.is_implementation_specific:
        raise AssertionError(
            f"(empwilli): "
            f"The class {cls.name!r} is implementation specific. "
            f"At the moment, we assume that all classes are not "
            f"implementation-specific, so that we can automatically generate the "
            f"shallow-equals methods. This way we can dispense of the whole "
            f"snippet/specific-implementation loading logic in "
            f"the unit test generation. Please notify the developers if you see this, "
            f"so that we can add the logic for implementation-specific classes "
            f"to this generation script."
        )

    cls_name = java_naming.class_name(cls.name)

    return Stripped(
        f"""\
private static Boolean {cls_name}DeepEquals({cls_name} that, {cls_name} other) {{
{I}return DeepEqualiserInstance.transform(that, other);
}}"""
    )


def generate(
    package: java_common.PackageIdentifier,
    symbol_table: intermediate.SymbolTable,
) -> List[java_common.JavaFile]:
    """
    Generate code to test copying.
    """
    blocks = [
        _generate_deep_equals_transformer(symbol_table=symbol_table),
        Stripped(
            """\
private static final DeepEqualiser DeepEqualiserInstance = new DeepEqualiser();"""
        ),
        Stripped(
            f"""\
/**
 * Compare two byte spans for equal content.
 */
private static Boolean byteSpansEqual(byte[] that, byte[] other) {{
{I}return that.equals(other);
}}"""
        ),
        Stripped(
            f"""\
private static class Pair<A, B> {{
{I}private final A first;
{I}private final B second;
{I}
{I}public Pair(A first, B second) {{
{II}this.first = first;
{II}this.second = second;
{I}}}
{I}
{I}public A getFirst() {{
{II}return first;
{I}}}
{I}
{I}public B getSecond() {{
{II}return second;
{I}}}
}}"""
        ),
        Stripped(
            f"""\
// Java 8 doesn't provide a zip operation out of the box, so we have to ship our own.
// Adapted from: https://stackoverflow.com/a/23529010
private static <A, B> Stream<Pair<A, B>> zip(
{I}Stream<? extends A> a,
{I}Stream<? extends B> b) {{
{I}Spliterator<? extends A> aSplit = Objects.requireNonNull(a).spliterator();
{I}Spliterator<? extends B> bSplit = Objects.requireNonNull(b).spliterator();
{I}
{I}int characteristics = aSplit.characteristics() & bSplit.characteristics() &
{II}~(Spliterator.DISTINCT | Spliterator.SORTED);
{I}
{I}long zipSize = ((characteristics & Spliterator.SIZED) != 0)
{II}? Math.min(aSplit.getExactSizeIfKnown(), bSplit.getExactSizeIfKnown())
{II}: -1;
{I}
{I}Iterator<A> aIter = Spliterators.iterator(aSplit);
{I}Iterator<B> bIter = Spliterators.iterator(bSplit);
{I}Iterator<Pair<A, B>> cIter = new Iterator<Pair<A, B>>() {{
{II}@Override
{II}public boolean hasNext() {{
{III}return aIter.hasNext() && bIter.hasNext();
{II}}}
{II}
{II}@Override
{II}public Pair<A, B> next() {{
{III}return new Pair<>(aIter.next(), bIter.next());
{II}}}
{I}}};
{I}
{I}Spliterator<Pair<A, B>> split = Spliterators.spliterator(cIter, zipSize, characteristics);
{I}return StreamSupport.stream(split, false);
}}"""
        ),
    ]  # type: List[Stripped]

    for concrete_cls in symbol_table.concrete_classes:
        blocks.append(_generate_shallow_equals(cls=concrete_cls))

    for concrete_cls in symbol_table.concrete_classes:
        blocks.append(_generate_deep_equals(cls=concrete_cls))

    for concrete_cls in symbol_table.concrete_classes:
        cls_name = java_naming.class_name(concrete_cls.name)

        blocks.append(
            Stripped(
                f"""\
@Test
public void test{cls_name}ShallowCopy() throws IOException {{
{I}final {cls_name} instance = CommonJsonization.loadMaximal{cls_name}();
{I}final {cls_name} instanceCopy = Copying.shallow(instance);

{I}assertTrue(
{II}{cls_name}ShallowEquals(instance, instanceCopy),
{II}{java_common.string_literal(cls_name)});
}} // public void test{cls_name}ShallowCopy"""
            )
        )

        blocks.append(
            Stripped(
                f"""\
@Test
public void test{cls_name}DeepCopy() throws IOException {{
{I}final {cls_name} instance = CommonJsonization.loadMaximal{cls_name}();
{I}final {cls_name} instanceCopy = Copying.deep(instance);

{I}assertTrue(
{II}{cls_name}DeepEquals(instance, instanceCopy),
{II}{java_common.string_literal(cls_name)});
}} // public void test{cls_name}DeepCopy"""
            )
        )

    blocks_joined = "\n\n".join(blocks)

    writer = io.StringIO()
    writer.write(
        f"""\
{java_common.WARNING}

package {package}.tests;

import static org.junit.jupiter.api.Assertions.assertTrue;

import {package}.copying.Copying;
import {package}.types.impl.*;
import {package}.types.model.*;
import {package}.types.model.IClass;
import {package}.visitation.AbstractTransformerWithContext;
import java.io.IOException;
import java.util.*;
import java.util.stream.Stream;
import java.util.stream.StreamSupport;
import org.junit.jupiter.api.Test;

public class TestCopying {{
{I}{indent_but_first_line(blocks_joined, I)}
}} // class TestCopying

// package {package}.tests

{java_common.WARNING}
"""
    )

    return [java_common.JavaFile("TestCopying.java", writer.getvalue())]


assert generate.__doc__ is not None
assert generate.__doc__.strip().startswith(__doc__.strip())



"""Generate code for XML de/serialization."""

import io
from typing import Tuple, Optional, List, Dict

from icontract import ensure, require

from aas_core_codegen import intermediate, naming, specific_implementations
from aas_core_codegen.common import (
    Error,
    Stripped,
    Identifier,
    assert_never,
    indent_but_first_line,
)
from aas_core_codegen.typescript import (
    common as typescript_common,
    naming as typescript_naming,
)
from aas_core_codegen.typescript.common import (
    INDENT as I,
    INDENT2 as II,
    INDENT3 as III,
    INDENT4 as IIII,
    INDENT5 as IIIII,
)


# region De-serialization


_PARSE_FUNCTION_BY_PRIMITIVE_TYPE = {
    intermediate.PrimitiveType.BOOL: Identifier("parseBooleanText"),
    intermediate.PrimitiveType.INT: Identifier("parseIntegerText"),
    intermediate.PrimitiveType.FLOAT: Identifier("parseFloatText"),
    intermediate.PrimitiveType.STR: Identifier("parseStringText"),
    intermediate.PrimitiveType.BYTEARRAY: Identifier("parseBase64EncodedBytesText"),
}


def _parse_function_for_atomic_type(
    type_annotation: intermediate.AtomicTypeAnnotation,
) -> Identifier:
    """Resolve the generated parse helper for an atomic XML text value."""
    if isinstance(type_annotation, intermediate.PrimitiveTypeAnnotation):
        return _PARSE_FUNCTION_BY_PRIMITIVE_TYPE[type_annotation.a_type]

    elif isinstance(type_annotation, intermediate.OurTypeAnnotation):
        our_type = type_annotation.our_type

        assert not isinstance(
            our_type, (intermediate.AbstractClass, intermediate.ConcreteClass)
        )

        if isinstance(our_type, intermediate.Enumeration):
            return typescript_naming.function_name(
                Identifier(f"parse_{our_type.name}_text")
            )

        elif isinstance(our_type, intermediate.ConstrainedPrimitive):
            return _PARSE_FUNCTION_BY_PRIMITIVE_TYPE[our_type.constrainee]

        else:
            assert_never(our_type)

    else:
        assert_never(type_annotation)


def _parse_function_name_for_concrete_class(
    cls: intermediate.ConcreteClass,
) -> Identifier:
    """Generate the name of the parse function for the concrete class."""
    return typescript_naming.function_name(
        Identifier(f"parse_{cls.name}_from_open_tag")
    )


def _generate_parse_text_as_enumeration(
    enumeration: intermediate.Enumeration,
) -> Stripped:
    """Generate parser for text representation of an enumeration literal."""
    enum_name = typescript_naming.enum_name(enumeration.name)
    parse_function_name = typescript_naming.function_name(
        Identifier(f"parse_{enumeration.name}_text")
    )
    from_string_function = typescript_naming.function_name(
        Identifier(f"{enumeration.name}_from_string")
    )

    return Stripped(
        f"""\
function {parse_function_name}(
{I}text: string
): AasCommon.Either<AasTypes.{enum_name}, DeserializationError> {{
{I}const literal = AasStringification.{from_string_function}(text);
{I}if (literal === null) {{
{II}return newDeserializationError<AasTypes.{enum_name}>(
{III}`Unexpected literal of {enum_name}: ${{text}}`
{II});
{I}}}

{I}return new AasCommon.Either<AasTypes.{enum_name}, DeserializationError>(
{II}literal,
{II}null
{I});
}}"""
    )


@require(lambda cls, prop: id(prop) in cls.property_id_set)
def _generate_parse_case_for_property(
    cls: intermediate.ConcreteClass,
    prop: intermediate.Property,
    var_name: Identifier,
) -> Stripped:
    """
    Generate a switch case to parse a property from XML element content.

    The generated code stores the parsed property value into ``var_name``.
    """
    xml_name_literal = typescript_common.string_literal(naming.xml_property(prop.name))

    type_anno = intermediate.beneath_optional(prop.type_annotation)

    duplicate_check = Stripped(
        f"""\
if ({var_name} !== null) {{
{I}propertyError = new DeserializationError(
{II}"Property " +
{III}{xml_name_literal} +
{III}" occurred more than once"
{I});
{I}break;
}}"""
    )

    if isinstance(
        type_anno,
        (intermediate.PrimitiveTypeAnnotation, intermediate.OurTypeAnnotation),
    ):
        if isinstance(type_anno, intermediate.OurTypeAnnotation) and isinstance(
            type_anno.our_type,
            (intermediate.AbstractClass, intermediate.ConcreteClass),
        ):
            our_type = type_anno.our_type
            as_function_name = typescript_naming.function_name(
                Identifier(f"as_{our_type.name}")
            )

            expected_name = typescript_naming.interface_name(our_type.name)

            if (
                isinstance(our_type, intermediate.ConcreteClass)
                and len(our_type.concrete_descendants) == 0
            ):
                parse_function_name = _parse_function_name_for_concrete_class(
                    cls=our_type
                )

                parse_body = Stripped(
                    f"""\
const classOrError = {parse_function_name}(cursor, propertyStartTag);
if (classOrError.error !== null) {{
{I}propertyError = classOrError.error;
{I}break;
}}

const casted = AasTypes.{as_function_name}(classOrError.mustValue());
if (casted === null) {{
{I}propertyError = new DeserializationError(
{II}"Expected property " +
{III}{xml_name_literal} +
{III}" to contain an instance of {expected_name}"
{I});
{I}break;
}}

{var_name} = casted;"""
                )
            else:
                parse_body = Stripped(
                    f"""\
const classOrError = parsePropertyAsClassInstance(cursor, propertyStartTag);
if (classOrError.error !== null) {{
{I}propertyError = classOrError.error;
{I}break;
}}

const casted = AasTypes.{as_function_name}(classOrError.mustValue());
if (casted === null) {{
{I}propertyError = new DeserializationError(
{II}"Expected property " +
{III}{xml_name_literal} +
{III}" to contain an instance of {expected_name}"
{I});
{I}break;
}}

{var_name} = casted;"""
                )
        else:
            parse_function = _parse_function_for_atomic_type(type_anno)
            parse_body = Stripped(
                f"""\
const textOrError = readTextContentAndConsumeEndTag(cursor, propertyStartTag);
if (textOrError.error !== null) {{
{I}propertyError = textOrError.error;
{I}break;
}}

const parsedOrError = {parse_function}(textOrError.mustValue());
if (parsedOrError.error !== null) {{
{I}propertyError = parsedOrError.error;
{I}break;
}}

{var_name} = parsedOrError.mustValue();"""
            )

    elif isinstance(type_anno, intermediate.ListTypeAnnotation):
        list_items_type = typescript_common.generate_type(
            type_anno.items,
            types_module=Identifier("AasTypes"),
        )

        if isinstance(
            type_anno.items,
            (intermediate.PrimitiveTypeAnnotation, intermediate.OurTypeAnnotation),
        ) and not (
            isinstance(type_anno.items, intermediate.OurTypeAnnotation)
            and isinstance(
                type_anno.items.our_type,
                (intermediate.AbstractClass, intermediate.ConcreteClass),
            )
        ):
            parse_item_function = _parse_function_for_atomic_type(type_anno.items)

            parse_body = Stripped(
                f"""\
const parsedItems = new Array<{list_items_type}>();
let itemIndex = 0;

cursor.skipIgnorable();
while (true) {{
{I}const itemOrClose = cursor.current();
{I}if (itemOrClose === null) {{
{II}propertyError = new DeserializationError(
{III}"Expected an XML element corresponding to a list item " +
{IIII}"or property closing element, but got end of token stream"
{II});
{II}break;
{I}}}

{I}if (itemOrClose instanceof CloseTagToken) {{
{II}const propertyCloseError = checkExpectedCloseTag(
{III}itemOrClose,
{III}localNameOfTag(propertyStartTag.tag)
{II});
{II}if (propertyCloseError !== null) {{
{III}propertyError = propertyCloseError;
{III}break;
{II}}}

{II}cursor.advance();
{II}break;
{I}}}

{I}if (!(itemOrClose instanceof OpenTagToken)) {{
{II}propertyError = new DeserializationError(
{III}"Expected an XML element corresponding to a list item " +
{IIII}`or property closing element, but got token kind: ${{itemOrClose.kind}}`
{II});
{II}break;
{I}}}

{I}const itemStartTag = itemOrClose;
{I}const itemNamespaceError = checkExpectedOpenTagNamespace(itemStartTag);
{I}if (itemNamespaceError !== null) {{
{II}propertyError = itemNamespaceError;
{II}break;
{I}}}

{I}const itemLocalName = localNameOfTag(itemStartTag.tag);
{I}if (itemLocalName !== "v") {{
{II}propertyError = new DeserializationError(
{III}`Expected list item element 'v', but got: ${{itemLocalName}}`
{II});
{II}break;
{I}}}

{I}cursor.advance();

{I}const itemTextOrError = readTextContentAndConsumeEndTag(cursor, itemStartTag);
{I}if (itemTextOrError.error !== null) {{
{II}propertyError = itemTextOrError.error;
{II}propertyError.path.prepend(new IndexSegment(itemIndex));
{II}break;
{I}}}

{I}const itemParsedOrError = {parse_item_function}(itemTextOrError.mustValue());
{I}if (itemParsedOrError.error !== null) {{
{II}propertyError = itemParsedOrError.error;
{II}propertyError.path.prepend(new IndexSegment(itemIndex));
{II}break;
{I}}}

{I}parsedItems.push(itemParsedOrError.mustValue());
{I}itemIndex++;
{I}cursor.skipIgnorable();
}}

if (propertyError === null) {{
{I}{var_name} = parsedItems;
}}"""
            )

        elif isinstance(type_anno.items, intermediate.OurTypeAnnotation) and isinstance(
            type_anno.items.our_type,
            (intermediate.AbstractClass, intermediate.ConcreteClass),
        ):
            as_function_name = typescript_naming.function_name(
                Identifier(f"as_{type_anno.items.our_type.name}")
            )
            expected_name = typescript_naming.interface_name(
                type_anno.items.our_type.name
            )

            parse_body = Stripped(
                f"""\
const parsedItems = new Array<{list_items_type}>();
let itemIndex = 0;

cursor.skipIgnorable();
while (true) {{
{I}const itemOrClose = cursor.current();
{I}if (itemOrClose === null) {{
{II}propertyError = new DeserializationError(
{III}"Expected an XML element corresponding to a list item " +
{IIII}"or property closing element, but got end of token stream"
{II});
{II}break;
{I}}}

{I}if (itemOrClose instanceof CloseTagToken) {{
{II}const propertyCloseError = checkExpectedCloseTag(
{III}itemOrClose,
{III}localNameOfTag(propertyStartTag.tag)
{II});
{II}if (propertyCloseError !== null) {{
{III}propertyError = propertyCloseError;
{III}break;
{II}}}

{II}cursor.advance();
{II}break;
{I}}}

{I}if (!(itemOrClose instanceof OpenTagToken)) {{
{II}propertyError = new DeserializationError(
{III}"Expected an XML element corresponding to a list item " +
{IIII}`or property closing element, but got token kind: ${{itemOrClose.kind}}`
{II});
{II}break;
{I}}}

{I}const itemStartTag = itemOrClose;
{I}const itemNamespaceError = checkExpectedOpenTagNamespace(itemStartTag);
{I}if (itemNamespaceError !== null) {{
{II}propertyError = itemNamespaceError;
{II}break;
{I}}}

{I}const dispatch = ROOT_DISPATCH_BY_LOCAL_NAME.get(localNameOfTag(itemStartTag.tag));
{I}if (dispatch === undefined) {{
{II}propertyError = new DeserializationError(
{III}`Unexpected XML element in list property: ${{localNameOfTag(itemStartTag.tag)}}`
{II});
{II}break;
{I}}}

{I}cursor.advance();

{I}const itemOrError = dispatch(cursor, itemStartTag);
{I}if (itemOrError.error !== null) {{
{II}propertyError = itemOrError.error;
{II}propertyError.path.prepend(new IndexSegment(itemIndex));
{II}break;
{I}}}

{I}const casted = AasTypes.{as_function_name}(itemOrError.mustValue());
{I}if (casted === null) {{
{II}propertyError = new DeserializationError(
{III}"Expected a list item instance of {expected_name}"
{II});
{II}propertyError.path.prepend(new IndexSegment(itemIndex));
{II}break;
{I}}}

{I}parsedItems.push(casted);
{I}itemIndex++;
{I}cursor.skipIgnorable();
}}

if (propertyError === null) {{
{I}{var_name} = parsedItems;
}}"""
            )
        else:
            parse_body = Stripped(
                f"""\
propertyError = new DeserializationError(
{I}"XML de-serialization of the list property " +
{II}{xml_name_literal} +
{II}" in {typescript_naming.class_name(cls.name)} can not handle list items " +
{II}"of this kind"
);"""
            )
    else:
        assert_never(type_anno)

    return Stripped(
        f"""\
case {xml_name_literal}: {{
{I}{indent_but_first_line(duplicate_check, I)}

{I}{indent_but_first_line(parse_body, I)}
{I}break;
}}"""
    )


def _generate_parse_concrete_class(cls: intermediate.ConcreteClass) -> Stripped:
    """Generate parser for a concrete class from a start XML tag."""
    function_name = _parse_function_name_for_concrete_class(cls=cls)
    cls_name = typescript_naming.class_name(cls.name)

    var_declarations = []  # type: List[Stripped]
    required_checks = []  # type: List[Stripped]
    parse_cases = []  # type: List[Stripped]

    var_name_by_property = {}  # type: Dict[Identifier, Identifier]

    for prop in cls.properties:
        var_name = typescript_naming.variable_name(Identifier(f"the_{prop.name}"))
        var_name_by_property[prop.name] = var_name

        var_type = typescript_common.generate_type(
            prop.type_annotation,
            types_module=Identifier("AasTypes"),
        )
        if not isinstance(prop.type_annotation, intermediate.OptionalTypeAnnotation):
            var_type = Stripped(f"{var_type} | null")

        var_declarations.append(Stripped(f"let {var_name}: {var_type} = null;"))

        if not isinstance(prop.type_annotation, intermediate.OptionalTypeAnnotation):
            message_literal = typescript_common.string_literal(
                f"The required property {naming.xml_property(prop.name)!r} is missing"
            )
            required_checks.append(
                Stripped(
                    f"""\
if ({var_name} === null) {{
{I}return newDeserializationError<AasTypes.Class>(
{II}{message_literal}
{I});
}}"""
                )
            )

        parse_cases.append(
            _generate_parse_case_for_property(cls=cls, prop=prop, var_name=var_name)
        )

    parse_cases.append(
        Stripped(
            """\
default: {
  propertyError = new DeserializationError(
    `Unexpected XML property: ${propertyLocalName}`
  );
  break;
}"""
        )
    )

    parse_cases_joined = "\n\n".join(parse_cases)

    if len(cls.constructor.arguments) == 0:
        construct = Stripped(
            f"""\
const instance = new AasTypes.{cls_name}();
return new AasCommon.Either<AasTypes.Class, DeserializationError>(
{I}instance,
{I}null
);"""
        )
    else:
        writer = io.StringIO()
        writer.write(f"const instance = new AasTypes.{cls_name}(\n")
        for i, arg in enumerate(cls.constructor.arguments):
            var_name = var_name_by_property[arg.name]
            writer.write(f"{I}{var_name}")
            if i < len(cls.constructor.arguments) - 1:
                writer.write(",\n")
            else:
                writer.write("\n")
        writer.write(
            f"""\
);
return new AasCommon.Either<AasTypes.Class, DeserializationError>(
{I}instance,
{I}null
);"""
        )
        construct = Stripped(writer.getvalue())

    declarations = (
        Stripped("\n".join(var_declarations))
        if len(var_declarations) > 0
        else Stripped("// No properties")
    )
    required_checks_block = (
        Stripped("\n\n".join(required_checks))
        if len(required_checks) > 0
        else Stripped("// No required properties")
    )

    return Stripped(
        f"""\
function {function_name}(
{I}cursor: XmlCursor,
{I}startTag: OpenTagToken
): AasCommon.Either<AasTypes.Class, DeserializationError> {{
{I}const observedLocalName = localNameOfTag(startTag.tag);
{I}{indent_but_first_line(declarations, I)}

{I}cursor.skipIgnorable();
{I}while (true) {{
{II}const token = cursor.current();
{II}if (token === null) {{
{III}return newDeserializationError<AasTypes.Class>(
{IIII}`Unexpected end of token stream while parsing {cls_name}`
{III});
{II}}}

{II}if (token instanceof CloseTagToken) {{
{III}break;
{II}}}

{II}if (!(token instanceof OpenTagToken)) {{
{III}return newDeserializationError<AasTypes.Class>(
{IIII}"Expected an XML property start element or the closing element of " +
{IIII}`{cls_name}, but got token kind: ${{token.kind}}`
{III});
{II}}}

{II}const namespaceError = checkExpectedOpenTagNamespace(token);
{II}if (namespaceError !== null) {{
{III}return new AasCommon.Either<AasTypes.Class, DeserializationError>(
{IIII}null,
{IIII}namespaceError
{III});
{II}}}

{II}const propertyStartTag = token;
{II}const propertyLocalName = localNameOfTag(propertyStartTag.tag);
{II}cursor.advance();

{II}let propertyError: DeserializationError | null = null;
{II}switch (propertyLocalName) {{
{III}{indent_but_first_line(parse_cases_joined, III)}
{II}}}

{II}if (propertyError !== null) {{
{III}propertyError.path.prepend(new NameSegment(propertyLocalName));
{III}return new AasCommon.Either<AasTypes.Class, DeserializationError>(
{IIII}null,
{IIII}propertyError
{III});
{II}}}

{II}cursor.skipIgnorable();
{I}}}

{I}const closeTag = cursor.current();
{I}if (!(closeTag instanceof CloseTagToken)) {{
{II}return newDeserializationError<AasTypes.Class>(
{III}"Expected closing element of " +
{IIII}`{cls_name}, but got token kind: ${{currentTokenKind(cursor)}}`
{II});
{I}}}

{I}const closeError = checkExpectedCloseTag(closeTag, observedLocalName);
{I}if (closeError !== null) {{
{II}return new AasCommon.Either<AasTypes.Class, DeserializationError>(
{III}null,
{III}closeError
{II});
{I}}}

{I}cursor.advance();

{I}{indent_but_first_line(required_checks_block, I)}

{I}{indent_but_first_line(construct, I)}
}}"""
    )


# endregion

# region Serialization


_SERIALIZE_FUNCTION_BY_PRIMITIVE_TYPE = {
    intermediate.PrimitiveType.BOOL: Identifier("serializeBooleanText"),
    intermediate.PrimitiveType.INT: Identifier("serializeIntegerText"),
    intermediate.PrimitiveType.FLOAT: Identifier("serializeFloatText"),
    intermediate.PrimitiveType.STR: Identifier("serializeStringText"),
    intermediate.PrimitiveType.BYTEARRAY: Identifier("serializeBase64EncodedBytesText"),
}


def _serialize_function_for_atomic_type(
    type_annotation: intermediate.AtomicTypeAnnotation,
) -> Identifier:
    """Resolve the name of the serialization function name for an atomic XML text value."""
    if isinstance(type_annotation, intermediate.PrimitiveTypeAnnotation):
        return _SERIALIZE_FUNCTION_BY_PRIMITIVE_TYPE[type_annotation.a_type]

    elif isinstance(type_annotation, intermediate.OurTypeAnnotation):
        our_type = type_annotation.our_type

        assert not isinstance(
            our_type, (intermediate.AbstractClass, intermediate.ConcreteClass)
        )

        if isinstance(our_type, intermediate.Enumeration):
            return typescript_naming.function_name(
                Identifier(f"serialize_{our_type.name}_text")
            )

        elif isinstance(our_type, intermediate.ConstrainedPrimitive):
            return _SERIALIZE_FUNCTION_BY_PRIMITIVE_TYPE[our_type.constrainee]

        else:
            assert_never(our_type)

    else:
        assert_never(type_annotation)


def _generate_serialize_text_as_enumeration(
    enumeration: intermediate.Enumeration,
) -> Stripped:
    """Generate serializer for text representation of an enumeration literal."""
    enum_name = typescript_naming.enum_name(enumeration.name)
    serialize_function_name = typescript_naming.function_name(
        Identifier(f"serialize_{enumeration.name}_text")
    )
    to_string_function = typescript_naming.function_name(
        Identifier(f"must_{enumeration.name}_to_string")
    )

    return Stripped(
        f"""\
function {serialize_function_name}(
{I}value: AasTypes.{enum_name}
): string {{
{I}return escapeXmlText(AasStringification.{to_string_function}(value));
}}"""
    )


def _generate_serialize_block_for_property(
    cls: intermediate.ConcreteClass,
    prop: intermediate.Property,
) -> Stripped:
    """Generate serialization statements for a property."""
    xml_name_literal = typescript_common.string_literal(naming.xml_property(prop.name))
    prop_name = typescript_naming.property_name(prop.name)
    access_expr = Stripped(f"that.{prop_name}")

    type_anno = intermediate.beneath_optional(prop.type_annotation)

    if isinstance(
        type_anno,
        (intermediate.PrimitiveTypeAnnotation, intermediate.OurTypeAnnotation),
    ):
        if isinstance(type_anno, intermediate.OurTypeAnnotation) and isinstance(
            type_anno.our_type,
            (intermediate.AbstractClass, intermediate.ConcreteClass),
        ):
            our_type = type_anno.our_type
            serialized_var = typescript_naming.variable_name(
                Identifier(f"serialized_{prop.name}")
            )
            if (
                isinstance(our_type, intermediate.ConcreteClass)
                and len(our_type.concrete_descendants) == 0
            ):
                body = Stripped(
                    f"""\
const {serialized_var} = this.transform({access_expr});
parts.push(openTag({xml_name_literal}));
parts.push({serialized_var}.innerXml);
parts.push(closeTag({xml_name_literal}));"""
                )
            else:
                body = Stripped(
                    f"""\
const {serialized_var} = this.transform({access_expr});
parts.push(openTag({xml_name_literal}));
parts.push(openTag({serialized_var}.localName));
parts.push({serialized_var}.innerXml);
parts.push(closeTag({serialized_var}.localName));
parts.push(closeTag({xml_name_literal}));"""
                )
        else:
            serialize_function = _serialize_function_for_atomic_type(type_anno)
            body = Stripped(
                f"""\
parts.push(openTag({xml_name_literal}));
parts.push({serialize_function}({access_expr}));
parts.push(closeTag({xml_name_literal}));"""
            )

    elif isinstance(type_anno, intermediate.ListTypeAnnotation):
        if isinstance(
            type_anno.items,
            (intermediate.PrimitiveTypeAnnotation, intermediate.OurTypeAnnotation),
        ) and not (
            isinstance(type_anno.items, intermediate.OurTypeAnnotation)
            and isinstance(
                type_anno.items.our_type,
                (intermediate.AbstractClass, intermediate.ConcreteClass),
            )
        ):
            serialize_item_function = _serialize_function_for_atomic_type(
                type_anno.items
            )
            item_var = typescript_naming.variable_name(Identifier(f"item_{prop.name}"))

            body = Stripped(
                f"""\
parts.push(openTag({xml_name_literal}));
for (const {item_var} of {access_expr}) {{
{I}parts.push(openTag("v"));
{I}parts.push({serialize_item_function}({item_var}));
{I}parts.push(closeTag("v"));
}}
parts.push(closeTag({xml_name_literal}));"""
            )

        elif isinstance(type_anno.items, intermediate.OurTypeAnnotation) and isinstance(
            type_anno.items.our_type,
            (intermediate.AbstractClass, intermediate.ConcreteClass),
        ):
            item_var = typescript_naming.variable_name(Identifier(f"item_{prop.name}"))
            serialized_item_var = typescript_naming.variable_name(
                Identifier(f"serialized_{prop.name}_item")
            )

            body = Stripped(
                f"""\
parts.push(openTag({xml_name_literal}));
for (const {item_var} of {access_expr}) {{
{I}const {serialized_item_var} = this.transform({item_var});
{I}parts.push(openTag({serialized_item_var}.localName));
{I}parts.push({serialized_item_var}.innerXml);
{I}parts.push(closeTag({serialized_item_var}.localName));
}}
parts.push(closeTag({xml_name_literal}));"""
            )

        else:
            body = Stripped(
                f"""\
throw new Error(
{I}"XML serialization of the list property " +
{II}{xml_name_literal} +
{II}" in {typescript_naming.class_name(cls.name)} can not handle list items " +
{II}"of this kind"
);"""
            )
    else:
        assert_never(type_anno)

    if isinstance(prop.type_annotation, intermediate.OptionalTypeAnnotation):
        return Stripped(
            f"""\
if ({access_expr} !== null) {{
{I}{indent_but_first_line(body, I)}
}}"""
        )

    return body


def _generate_transform_of_concrete_class(cls: intermediate.ConcreteClass) -> Stripped:
    """Generate ``transformX`` to serialize a concrete class to XML parts."""
    method_name = typescript_naming.method_name(Identifier(f"transform_{cls.name}"))
    cls_name = typescript_naming.class_name(cls.name)
    local_name_literal = typescript_common.string_literal(
        naming.xml_class_name(cls.name)
    )

    blocks = [Stripped("const parts = new Array<string>();")]  # type: List[Stripped]

    for prop in cls.properties:
        blocks.append(_generate_serialize_block_for_property(cls=cls, prop=prop))

    blocks.append(
        Stripped(
            f"""\
return {{
{I}localName: {local_name_literal},
{I}innerXml: parts.join("")
}};"""
        )
    )

    writer = io.StringIO()
    writer.write(
        f"""\
/**
 * Serialize `that` to an XML element representation.
 *
 * @param that - instance to be serialized
 * @returns serialized XML element representation
 */
{method_name}(
{I}that: AasTypes.{cls_name}
): SerializedElement {{
"""
    )

    for i, block in enumerate(blocks):
        if i > 0:
            writer.write("\n\n")
        writer.write(indent_but_first_line(block, I))

    writer.write("\n}")
    return Stripped(writer.getvalue())


def _generate_serializer(symbol_table: intermediate.SymbolTable) -> Stripped:
    """Generate the serializer transformer over all concrete classes."""
    methods = []  # type: List[Stripped]

    for cls in symbol_table.concrete_classes:
        methods.append(_generate_transform_of_concrete_class(cls=cls))

    writer = io.StringIO()
    writer.write(
        """\
/**
 * Serialize an AAS instance to XML parts.
 */
class Serializer extends AasTypes.AbstractTransformer<SerializedElement> {
"""
    )

    for method in methods:
        writer.write("\n\n")
        writer.write(indent_but_first_line(method, I))

    writer.write("\n}")

    return Stripped(writer.getvalue())


def _generate_root_dispatch_map(symbol_table: intermediate.SymbolTable) -> Stripped:
    """Generate the dispatch map from root XML local names to parse functions."""
    writer = io.StringIO()
    writer.write(
        f"""\
const ROOT_DISPATCH_BY_LOCAL_NAME =
{I}new Map<
{II}string,
{II}(
{III}cursor: XmlCursor,
{III}startTag: OpenTagToken
{II}) => AasCommon.Either<AasTypes.Class, DeserializationError>
{I}>([
"""
    )

    for i, cls in enumerate(symbol_table.concrete_classes):
        local_name_literal = typescript_common.string_literal(
            naming.xml_class_name(cls.name)
        )
        parse_function_name = _parse_function_name_for_concrete_class(cls=cls)

        writer.write(
            f"""\
{II}[
{III}{local_name_literal},
{III}{parse_function_name}
{II}]"""
        )

        if i < len(symbol_table.concrete_classes) - 1:
            writer.write(",\n")
        else:
            writer.write("\n")

    writer.write(
        f"""\
{I}]);"""
    )

    return Stripped(writer.getvalue())


# endregion


# fmt: off
@ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
@ensure(
    lambda result:
    not (result[0] is not None) or result[0].endswith('\n'),
    "Trailing newline mandatory for valid end-of-files"
)
# fmt: on
def generate(
    symbol_table: intermediate.SymbolTable,
    spec_impls: specific_implementations.SpecificImplementations,
) -> Tuple[Optional[str], Optional[List[Error]]]:
    """Generate code for XML de/serialization."""
    del spec_impls

    namespace_literal = typescript_common.string_literal(
        symbol_table.meta_model.xml_namespace
    )

    blocks = [
        Stripped(
            """\
/**
 * Provide de/serialization of AAS classes to/from XML.
 *
 * The implementation is incremental and follows a SAX-style parsing approach.
 */"""
        ),
        typescript_common.WARNING,
        Stripped(
            """\
import * as AasCommon from "./common";
import * as AasTypes from "./types";
import * as AasStringification from "./stringification";

import {
  CdataToken,
  CloseTagToken,
  CommentToken,
  EndToken,
  OpenTagToken,
  TextToken,
  XmlAnyToken,
  XmlSaxParser
} from "xmlsax-typescript";"""
        ),
        Stripped(
            f"""\
const NAMESPACE = {namespace_literal};"""
        ),
        Stripped(
            f"""\
/**
 * Represent a property name segment in an XML path.
 */
export class NameSegment {{
{I}readonly name: string;

{I}constructor(name: string) {{
{II}this.name = name;
{I}}}
}}"""
        ),
        Stripped(
            f"""\
/**
 * Represent an index segment in an XML path.
 */
export class IndexSegment {{
{I}readonly index: number;

{I}constructor(index: number) {{
{II}this.index = index;
{I}}}
}}"""
        ),
        Stripped(
            """\
export type Segment = NameSegment | IndexSegment;"""
        ),
        Stripped(
            f"""\
/**
 * Represent a relative path to the erroneous XML value.
 */
export class Path {{
{I}private readonly _segments = new Array<Segment>();

{I}segments(): Array<Segment> {{
{II}return this._segments;
{I}}}

{I}prepend(segment: Segment): void {{
{II}this._segments.unshift(segment);
{I}}}

{I}toString(): string {{
{II}if (this._segments.length === 0) {{
{III}return "";
{II}}}

{II}const parts = new Array<string>();
{II}for (const segment of this._segments) {{
{III}if (segment instanceof NameSegment) {{
{IIII}if (parts.length === 0) {{
{IIIII}parts.push(segment.name);
{IIII}}} else {{
{IIIII}parts.push(`.${{segment.name}}`);
{IIII}}}
{III}}} else if (segment instanceof IndexSegment) {{
{IIII}parts.push(`[${{segment.index}}]`);
{III}}}
{II}}}

{II}return parts.join("");
{I}}}
}}"""
        ),
        Stripped(
            f"""\
/**
 * Signal that XML de-serialization could not be performed.
 */
export class DeserializationError {{
{I}readonly message: string;
{I}readonly path: Path;

{I}constructor(message: string, path: Path | null = null) {{
{II}this.message = message;
{II}this.path = path ?? new Path();
{I}}}
}}"""
        ),
        Stripped(
            f"""\
/**
 * Signal that XML serialization could not be performed.
 */
export class SerializationError {{
{I}readonly message: string;
{I}readonly path: Path;

{I}constructor(message: string, path: Path | null = null) {{
{II}this.message = message;
{II}this.path = path ?? new Path();
{I}}}
}}"""
        ),
        Stripped(
            f"""\
function newDeserializationError<T>(
{I}message: string
): AasCommon.Either<T, DeserializationError> {{
{I}return new AasCommon.Either<T, DeserializationError>(
{II}null,
{II}new DeserializationError(message)
{I});
}}

function currentTokenKind(cursor: XmlCursor): string {{
{I}const token = cursor.current();
{I}if (token === null) {{
{II}return "end-of-token-stream";
{I}}}

{I}return token.kind;
}}

function localNameOfTag(tag: unknown): string {{
{I}const aTag = tag as {{
{II}name?: unknown,
{II}local?: unknown,
{II}localName?: unknown
{I}}};

{I}if (typeof aTag.local === "string") {{
{II}return aTag.local;
{I}}}
{I}if (typeof aTag.localName === "string") {{
{II}return aTag.localName;
{I}}}
{I}if (typeof aTag.name === "string") {{
{II}const colonIndex = aTag.name.indexOf(":");
{II}if (colonIndex >= 0) {{
{III}return aTag.name.substring(colonIndex + 1);
{II}}}
{II}return aTag.name;
{I}}}

{I}return "";
}}

function namespaceOfTag(tag: unknown): string {{
{I}const aTag = tag as {{ uri?: unknown, namespaceURI?: unknown }};
{I}if (typeof aTag.uri === "string") {{
{II}return aTag.uri;
{I}}}
{I}if (typeof aTag.namespaceURI === "string") {{
{II}return aTag.namespaceURI;
{I}}}
{I}return "";
}}

function checkExpectedOpenTagNamespace(
{I}openTag: OpenTagToken
): DeserializationError | null {{
{I}const namespace = namespaceOfTag(openTag.tag);
{I}if (namespace !== NAMESPACE) {{
{II}return new DeserializationError(
{III}"Expected XML namespace " +
{IIII}`'${{NAMESPACE}}', but got '${{namespace}}'`
{II});
{I}}}

{I}return null;
}}

function checkExpectedCloseTag(
{I}closeTag: CloseTagToken,
{I}expectedLocalName: string
): DeserializationError | null {{
{I}const namespace = namespaceOfTag(closeTag.tag);
{I}if (namespace !== NAMESPACE) {{
{II}return new DeserializationError(
{III}"Expected XML namespace " +
{IIII}`'${{NAMESPACE}}', but got '${{namespace}}'`
{II});
{I}}}

{I}const observedLocalName = localNameOfTag(closeTag.tag);
{I}if (observedLocalName !== expectedLocalName) {{
{II}return new DeserializationError(
{III}`Expected closing XML element '${{expectedLocalName}}', ` +
{III}`but got '${{observedLocalName}}'`
{II});
{I}}}

{I}return null;
}}

/**
 * Cursor over parsed XML SAX tokens.
 */
class XmlCursor {{
{I}private readonly _tokens: Array<XmlAnyToken>;
{I}private _index = 0;

{I}constructor(tokens: Array<XmlAnyToken>) {{
{II}this._tokens = tokens;
{I}}}

{I}current(): XmlAnyToken | null {{
{II}if (this._index >= this._tokens.length) {{
{III}return null;
{II}}}
{II}return this._tokens[this._index];
{I}}}

{I}advance(): void {{
{II}if (this._index < this._tokens.length) {{
{III}this._index++;
{II}}}
{I}}}

{I}skipIgnorable(): void {{
{II}while (true) {{
{III}const token = this.current();
{III}if (token === null) {{
{IIII}break;
{III}}}

{III}if (token instanceof CommentToken) {{
{IIII}this.advance();
{IIII}continue;
{III}}}

{III}if (token instanceof TextToken || token instanceof CdataToken) {{
{IIII}if (token.text.trim().length === 0) {{
{IIIII}this.advance();
{IIIII}continue;
{IIII}}}
{III}}}

{III}break;
{II}}}
{I}}}
}}

function tokenizeXml(
{I}xml: string
): AasCommon.Either<Array<XmlAnyToken>, DeserializationError> {{
{I}const parser = new XmlSaxParser({{ allowDoctype: false, xmlns: true }});
{I}const tokens = new Array<XmlAnyToken>();

{I}try {{
{II}for (const token of parser.feed(xml)) {{
{III}tokens.push(token);
{II}}}
{II}for (const token of parser.close()) {{
{III}if (!(token instanceof EndToken)) {{
{IIII}tokens.push(token);
{III}}}
{II}}}
{I}}} catch (error) {{
{II}return newDeserializationError<Array<XmlAnyToken>>(
{III}`Failed to parse XML: ${{error}}`
{II});
{I}}}

{I}return new AasCommon.Either<Array<XmlAnyToken>, DeserializationError>(
{II}tokens,
{II}null
{I});
}}

function readRequiredRootOpenTag(
{I}cursor: XmlCursor
): AasCommon.Either<OpenTagToken, DeserializationError> {{
{I}cursor.skipIgnorable();

{I}const token = cursor.current();
{I}if (token === null) {{
{II}return newDeserializationError<OpenTagToken>(
{III}"Expected a root XML element, but got an empty token stream"
{II});
{I}}}

{I}if (!(token instanceof OpenTagToken)) {{
{II}return newDeserializationError<OpenTagToken>(
{III}`Expected a root XML start element, but got token kind: ${{token.kind}}`
{II});
{I}}}

{I}const namespaceError = checkExpectedOpenTagNamespace(token);
{I}if (namespaceError !== null) {{
{II}return new AasCommon.Either<OpenTagToken, DeserializationError>(
{III}null,
{III}namespaceError
{II});
{I}}}

{I}cursor.advance();

{I}return new AasCommon.Either<OpenTagToken, DeserializationError>(
{II}token,
{II}null
{I});
}}

function readTextContentAndConsumeEndTag(
{I}cursor: XmlCursor,
{I}startTag: OpenTagToken
): AasCommon.Either<string, DeserializationError> {{
{I}cursor.skipIgnorable();

{I}let text = "";
{I}const maybeText = cursor.current();
{I}if (maybeText instanceof TextToken || maybeText instanceof CdataToken) {{
{II}text = maybeText.text;
{II}cursor.advance();
{II}cursor.skipIgnorable();
{I}}}

{I}const closeToken = cursor.current();
{I}if (!(closeToken instanceof CloseTagToken)) {{
{II}return newDeserializationError<string>(
{III}"Expected property closing XML element, but got token kind: " +
{IIII}currentTokenKind(cursor)
{II});
{I}}}

{I}const localName = localNameOfTag(startTag.tag);
{I}const closeError = checkExpectedCloseTag(closeToken, localName);
{I}if (closeError !== null) {{
{II}return new AasCommon.Either<string, DeserializationError>(
{III}null,
{III}closeError
{II});
{I}}}

{I}cursor.advance();

{I}return new AasCommon.Either<string, DeserializationError>(
{II}text,
{II}null
{I});
}}

function parseBooleanText(
{I}text: string
): AasCommon.Either<boolean, DeserializationError> {{
{I}if (text === "true" || text === "1") {{
{II}return new AasCommon.Either<boolean, DeserializationError>(true, null);
{I}}}
{I}if (text === "false" || text === "0") {{
{II}return new AasCommon.Either<boolean, DeserializationError>(false, null);
{I}}}

{I}return newDeserializationError<boolean>(
{II}`Expected xs:boolean text, but got: ${{text}}`
{I});
}}

function parseIntegerText(
{I}text: string
): AasCommon.Either<number, DeserializationError> {{
{I}if (!/^[+-]?\\d+$/.test(text)) {{
{II}return newDeserializationError<number>(
{III}`Expected integer text, but got: ${{text}}`
{II});
{I}}}

{I}const value = Number(text);
{I}if (!Number.isInteger(value)) {{
{II}return newDeserializationError<number>(
{III}`Expected integer text, but got: ${{text}}`
{II});
{I}}}

{I}return new AasCommon.Either<number, DeserializationError>(value, null);
}}

function parseFloatText(
{I}text: string
): AasCommon.Either<number, DeserializationError> {{
{I}if (text === "INF") {{
{II}return new AasCommon.Either<number, DeserializationError>(Infinity, null);
{I}}}
{I}if (text === "-INF") {{
{II}return new AasCommon.Either<number, DeserializationError>(-Infinity, null);
{I}}}
{I}if (text === "NaN") {{
{II}return new AasCommon.Either<number, DeserializationError>(NaN, null);
{I}}}

{I}const value = Number(text);
{I}if (Number.isNaN(value)) {{
{II}return newDeserializationError<number>(
{III}`Expected xs:double text, but got: ${{text}}`
{II});
{I}}}

{I}return new AasCommon.Either<number, DeserializationError>(value, null);
}}

function parseStringText(
{I}text: string
): AasCommon.Either<string, DeserializationError> {{
{I}return new AasCommon.Either<string, DeserializationError>(text, null);
}}

function parseBase64EncodedBytesText(
{I}text: string
): AasCommon.Either<Uint8Array, DeserializationError> {{
{I}const decodedOrError = AasCommon.base64Decode(text);
{I}if (decodedOrError.error !== null) {{
{II}return newDeserializationError<Uint8Array>(
{III}decodedOrError.error
{II});
{I}}}

{I}return new AasCommon.Either<Uint8Array, DeserializationError>(
{II}decodedOrError.mustValue(),
{II}null
{I});
}}

function parsePropertyAsClassInstance(
{I}cursor: XmlCursor,
{I}propertyStartTag: OpenTagToken
): AasCommon.Either<AasTypes.Class, DeserializationError> {{
{I}cursor.skipIgnorable();

{I}const token = cursor.current();
{I}if (!(token instanceof OpenTagToken)) {{
{II}return newDeserializationError<AasTypes.Class>(
{III}"Expected nested class element in XML property, but got token kind: " +
{IIII}currentTokenKind(cursor)
{II});
{I}}}

{I}const namespaceError = checkExpectedOpenTagNamespace(token);
{I}if (namespaceError !== null) {{
{II}return new AasCommon.Either<AasTypes.Class, DeserializationError>(
{III}null,
{III}namespaceError
{II});
{I}}}

{I}const localName = localNameOfTag(token.tag);
{I}const dispatch = ROOT_DISPATCH_BY_LOCAL_NAME.get(localName);
{I}if (dispatch === undefined) {{
{II}return newDeserializationError<AasTypes.Class>(
{III}`Unexpected nested class XML element: ${{localName}}`
{II});
{I}}}

{I}cursor.advance();
{I}const instanceOrError = dispatch(cursor, token);
{I}if (instanceOrError.error !== null) {{
{II}return instanceOrError;
{I}}}

{I}cursor.skipIgnorable();
{I}const propertyCloseToken = cursor.current();
{I}if (!(propertyCloseToken instanceof CloseTagToken)) {{
{II}return newDeserializationError<AasTypes.Class>(
{III}"Expected property closing XML element after nested class, but got token kind: " +
{IIII}currentTokenKind(cursor)
{II});
{I}}}

{I}const expectedPropertyLocalName = localNameOfTag(propertyStartTag.tag);
{I}const propertyCloseError = checkExpectedCloseTag(
{II}propertyCloseToken,
{II}expectedPropertyLocalName
{I});
{I}if (propertyCloseError !== null) {{
{II}return new AasCommon.Either<AasTypes.Class, DeserializationError>(
{III}null,
{III}propertyCloseError
{II});
{I}}}

{I}cursor.advance();

{I}return instanceOrError;
}}"""
        ),
    ]  # type: List[Stripped]

    for enumeration in symbol_table.enumerations:
        blocks.append(_generate_parse_text_as_enumeration(enumeration))

    for enumeration in symbol_table.enumerations:
        blocks.append(_generate_serialize_text_as_enumeration(enumeration))

    for concrete_cls in symbol_table.concrete_classes:
        blocks.append(_generate_parse_concrete_class(cls=concrete_cls))

    blocks.extend(
        [
            _generate_root_dispatch_map(symbol_table=symbol_table),
            Stripped(
                f"""\
/**
 * Parse an XML string as an AAS instance.
 *
 * @param xml - XML string to parse
 * @returns parsed AAS instance or an error
 */
export function fromXmlString(
{I}xml: string
): AasCommon.Either<AasTypes.Class, DeserializationError> {{
{I}if (xml.length === 0) {{
{II}return newDeserializationError<AasTypes.Class>(
{III}"Expected an XML document, but got an empty string"
{II});
{I}}}

{I}const tokensOrError = tokenizeXml(xml);
{I}if (tokensOrError.error !== null) {{
{II}return new AasCommon.Either<AasTypes.Class, DeserializationError>(
{III}null,
{III}tokensOrError.error
{II});
{I}}}

{I}const cursor = new XmlCursor(tokensOrError.mustValue());

{I}const rootOpenTagOrError = readRequiredRootOpenTag(cursor);
{I}if (rootOpenTagOrError.error !== null) {{
{II}return new AasCommon.Either<AasTypes.Class, DeserializationError>(
{III}null,
{III}rootOpenTagOrError.error
{II});
{I}}}

{I}const rootOpenTag = rootOpenTagOrError.mustValue();
{I}const rootLocalName = localNameOfTag(rootOpenTag.tag);

{I}const dispatch = ROOT_DISPATCH_BY_LOCAL_NAME.get(rootLocalName);
{I}if (dispatch === undefined) {{
{II}return newDeserializationError<AasTypes.Class>(
{III}`Unexpected root XML element: ${{rootLocalName}}`
{II});
{I}}}

{I}const instanceOrError = dispatch(cursor, rootOpenTag);
{I}if (instanceOrError.error !== null) {{
{II}return instanceOrError;
{I}}}

{I}cursor.skipIgnorable();
{I}const tokenAfterInstance = cursor.current();
{I}if (tokenAfterInstance !== null) {{
{II}if (!(tokenAfterInstance instanceof CloseTagToken)) {{
{III}return newDeserializationError<AasTypes.Class>(
{IIII}"Expected root closing XML element, but got token kind: " +
{IIII}currentTokenKind(cursor)
{III});
{II}}}

{II}const closeError = checkExpectedCloseTag(
{III}tokenAfterInstance,
{III}rootLocalName
{II});
{II}if (closeError !== null) {{
{III}return new AasCommon.Either<AasTypes.Class, DeserializationError>(
{IIII}null,
{IIII}closeError
{III});
{II}}}

{II}cursor.advance();
{I}}}

{I}cursor.skipIgnorable();
{I}if (cursor.current() !== null) {{
{II}return newDeserializationError<AasTypes.Class>(
{III}"Expected no tokens after the root XML element, but got token kind: " +
{IIII}currentTokenKind(cursor)
{II});
{I}}}

{I}return instanceOrError;
}}"""
            ),
            Stripped(
                f"""\
type SerializedElement = {{
{I}localName: string;
{I}innerXml: string;
}};

function openTag(localName: string, withNamespace = false): string {{
{I}if (withNamespace) {{
{II}return `<${{localName}} xmlns="${{NAMESPACE}}">`;
{I}}}

{I}return `<${{localName}}>`;
}}

function closeTag(localName: string): string {{
{I}return `</${{localName}}>`;
}}

function escapeXmlText(text: string): string {{
{I}return text
{II}.replace(/&/g, "&amp;")
{II}.replace(/</g, "&lt;")
{II}.replace(/>/g, "&gt;")
{II}.replace(/\"/g, "&quot;")
{II}.replace(/'/g, "&apos;");
}}

function serializeBooleanText(value: boolean): string {{
{I}return value ? "true" : "false";
}}

function serializeIntegerText(value: number): string {{
{I}if (!Number.isInteger(value)) {{
{II}throw new Error(`Expected an integer, but got: ${{value}}`);
{I}}}

{I}return `${{value}}`;
}}

function serializeFloatText(value: number): string {{
{I}if (Number.isNaN(value)) {{
{II}return "NaN";
{I}}}
{I}if (value === Infinity) {{
{II}return "INF";
{I}}}
{I}if (value === -Infinity) {{
{II}return "-INF";
{I}}}

{I}return `${{value}}`;
}}

function serializeStringText(value: string): string {{
{I}return escapeXmlText(value);
}}

function serializeBase64EncodedBytesText(value: Uint8Array): string {{
{I}return escapeXmlText(AasCommon.base64Encode(value));
}}"""
            ),
            _generate_serializer(symbol_table=symbol_table),
            Stripped("const SERIALIZER = new Serializer();"),
            Stripped(
                f"""\
/**
 * Serialize an AAS instance as an XML string.
 *
 * @param that - AAS instance to serialize
 * @returns serialized XML string
 */
export function toXmlString(that: AasTypes.Class): string {{
{I}const serialized = SERIALIZER.transform(that);
{I}const parts = new Array<string>();
{I}parts.push(openTag(serialized.localName, true));
{I}parts.push(serialized.innerXml);
{I}parts.push(closeTag(serialized.localName));
{I}return parts.join("");
}}"""
            ),
            typescript_common.WARNING,
        ]
    )

    writer = io.StringIO()
    for i, block in enumerate(blocks):
        if i > 0:
            writer.write("\n\n")

        writer.write(block)

    writer.write("\n")

    return writer.getvalue(), None


assert generate.__doc__ is not None
assert generate.__doc__.strip().startswith(__doc__.strip())



"""Provide generators for the main library."""

from aas_core_codegen.cpp.lib import (
    _generate_common,
    _generate_constants,
    _generate_enhancing,
    _generate_iteration,
    _generate_jsonization,
    _generate_pattern,
    _generate_revm,
    _generate_stringification,
    _generate_types,
    _generate_verification,
    _generate_visitation,
    _generate_wstringification,
    _generate_xmlization,
)

generate_common_header = _generate_common.generate_header
generate_common_implementation = _generate_common.generate_implementation

generate_constants_header = _generate_constants.generate_header
generate_constants_implementation = _generate_constants.generate_implementation

generate_enhancing_header = _generate_enhancing.generate_header

generate_iteration_header = _generate_iteration.generate_header
generate_iteration_implementation = _generate_iteration.generate_implementation

generate_jsonization_header = _generate_jsonization.generate_header
generate_jsonization_implementation = _generate_jsonization.generate_implementation

generate_pattern_header = _generate_pattern.generate_header
generate_pattern_implementation = _generate_pattern.generate_implementation

generate_revm_header = _generate_revm.generate_header
generate_revm_implementation = _generate_revm.generate_implementation

generate_stringification_header = _generate_stringification.generate_header
generate_stringification_implementation = (
    _generate_stringification.generate_implementation
)

generate_types_header = _generate_types.generate_header
generate_types_implementation = _generate_types.generate_implementation

generate_verification_header = _generate_verification.generate_header
generate_verification_implementation = _generate_verification.generate_implementation

generate_visitation_header = _generate_visitation.generate_header
generate_visitation_implementation = _generate_visitation.generate_implementation

generate_wstringification_header = _generate_wstringification.generate_header
generate_wstringification_implementation = (
    _generate_wstringification.generate_implementation
)

generate_xmlization_header = _generate_xmlization.generate_header
generate_xmlization_implementation = _generate_xmlization.generate_implementation

verify_for_types = _generate_types.verify



"""Generate code to test the ``Descend`` methods and ``VisitorThrough``."""

from typing import List

from aas_core_codegen import intermediate, naming
from aas_core_codegen.common import Stripped, indent_but_first_line
from aas_core_codegen.java import common as java_common, naming as java_naming
from aas_core_codegen.java.common import (
    INDENT as I,
    INDENT2 as II,
    INDENT3 as III,
    INDENT4 as IIII,
    INDENT5 as IIIII,
)


def generate(
    package: java_common.PackageIdentifier,
    symbol_table: intermediate.SymbolTable,
) -> List[java_common.JavaFile]:
    """
    Generate code to test the ``Descend`` methods and ``VisitorThrough``.
    """
    blocks = [
        Stripped(
            f"""\
private class TracingVisitorThrough extends VisitorThrough {{
{I}public final List<String> log = new ArrayList<>();

{I}@Override
{I}public void visit(IClass that) {{
{II}log.add(trace(that));
{II}super.visit(that);
{I}}}
}}"""
        ),
        Stripped(
            f"""\
private String trace(IClass instance) {{
{I}if (instance instanceof IIdentifiable) {{
{II}return instance.getClass().getSimpleName() + " with ID " + (((IIdentifiable) instance).getId());
{I}}} else if (instance instanceof IReferable) {{
{II}return instance.getClass().getSimpleName() + " with ID-short " + (((IReferable) instance).getIdShort());
{I}}} else {{
{II}return instance.getClass().getSimpleName();
{I}}}
        }}"""
        ),
        Stripped(
            f"""\
private void assertDescendAndVisitorThroughSame(IClass instance)
{{
{I}final List<String> logFromDescend = new ArrayList<>();

{I}for (IClass subInstance : instance.descend()) {{
{II}logFromDescend.add(trace(subInstance));
{I}}}

{I}final TracingVisitorThrough visitor = new TracingVisitorThrough();
{I}visitor.visit(instance);
{I}final List<String> traceFromVisitor = visitor.log;

{I}assertFalse(traceFromVisitor.isEmpty());

{I}assertEquals(trace(instance), traceFromVisitor.get(0));

{I}traceFromVisitor.remove(0);

{I}assertTrue(traceFromVisitor.equals(logFromDescend));
}}"""
        ),
        Stripped(
            f"""\
private void compareOrRerecordTrace(IClass instance, Path expectedPath) throws IOException {{
{I}final StringBuilder stringBuilder = new StringBuilder();
{I}for (IClass descendant : instance.descend()) {{
{II}stringBuilder.append(Common.trace(descendant));
{I}}}

{I}final String got = stringBuilder.toString();
{I}if (Common.RECORD_MODE) {{
{II}Files.createDirectories(expectedPath.getParent());
{II}Files.write(expectedPath, got.getBytes());
{I}}} else {{
{II}if (!Files.exists(expectedPath)) {{
{III}throw new FileNotFoundException(
{IIIII}"The file with the recorded value does not exist: " + expectedPath);
{II}}}
{II}final String expected =
{IIII}Files.readAllLines(expectedPath).stream().collect(Collectors.joining("\\n"));
{II}assertEquals(expected.replace("\\n", ""), got.replace("\\n", ""));
{I}}}
}}"""
        ),
    ]  # type: List[str]

    for concrete_cls in symbol_table.concrete_classes:
        cls_name_java = java_naming.class_name(concrete_cls.name)
        cls_name_json = naming.json_model_type(concrete_cls.name)

        blocks.append(
            Stripped(
                f"""\
@Test
public void testDescendOf{cls_name_java}() throws IOException {{
{I}final {cls_name_java} instance = CommonJsonization.loadMaximal{cls_name_java}();

{I}compareOrRerecordTrace(
{II}instance,
{II}Paths.get(
{III}Common.TEST_DATA_DIR,
{III}"Descend",
{III}{java_common.string_literal(cls_name_json)},
{III}"maximal.json.trace"));
}} // public void testDescendOf{cls_name_java}

@Test
public void testDescendAgainstVisitorThroughFor{cls_name_java}() throws IOException {{
{I}{cls_name_java} instance = (
{II}CommonJsonization.loadMaximal{cls_name_java}());

{I}assertDescendAndVisitorThroughSame(instance);
}} // public void testDescendAgainstVisitorThroughFor{cls_name_java}"""
            )
        )

    blocks_joined = "\n\n".join(blocks)

    return [
        java_common.JavaFile(
            "TestDescendAndVisitorThrough.java",
            f"""\
{java_common.WARNING}

package {package}.tests;

import static org.junit.jupiter.api.Assertions.*;

import {package}.types.impl.*;
import {package}.types.model.IClass;
import {package}.types.model.IIdentifiable;
import {package}.types.model.IReferable;
import {package}.visitation.VisitorThrough;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.ArrayList;
import java.util.List;
import java.util.stream.Collectors;
import org.junit.jupiter.api.Test;

public class TestDescendAndVisitorThrough {{
{I}{indent_but_first_line(blocks_joined, II)}
}} // class TestDescendAndVisitorThrough

// package {package}.tests

{java_common.WARNING}
""",
        )
    ]


assert generate.__doc__ is not None
assert generate.__doc__.strip().startswith(__doc__.strip())



"""Generate code to test the ``over{Enum}`` methods."""


import io
from typing import List

from icontract import ensure

from aas_core_codegen import intermediate
from aas_core_codegen.common import Stripped, Identifier, indent_but_first_line
from aas_core_codegen.typescript import (
    common as typescript_common,
    naming as typescript_naming,
)
from aas_core_codegen.typescript.common import (
    INDENT as I,
)


# fmt: off
@ensure(
    lambda result: result.endswith('\n'),
    "Trailing newline mandatory for valid end-of-files"
)
# fmt: on
def generate(symbol_table: intermediate.SymbolTable) -> str:
    """Generate code to test the ``over{Enum}`` methods."""
    blocks = [
        Stripped(
            """\
/**
 * Test `over{Enum}` functions.
 */"""
        ),
        typescript_common.WARNING,
        Stripped('import * as AasTypes from "../src/types";'),
    ]  # type: List[Stripped]

    for enumeration in symbol_table.enumerations:
        name = typescript_naming.enum_name(enumeration.name)

        body_writer = io.StringIO()
        body_writer.write(
            f"""\
const expected: Array<AasTypes.{name}> = [
"""
        )
        for i, literal in enumerate(enumeration.literals):
            literal_name = typescript_naming.enum_literal_name(literal.name)
            body_writer.write(f"{I}AasTypes.{name}.{literal_name}")

            if i < len(enumeration.literals) - 1:
                body_writer.write(",")

            body_writer.write("\n")

        body_writer.write("];\n\n")

        over_enum_name = typescript_naming.function_name(
            Identifier(f"over_{enumeration.name}")
        )

        body_writer.write(
            f"""\
const got = new Array<AasTypes.{name}>();
for (const literal of AasTypes.{over_enum_name}()) {{
{I}got.push(literal);
}}

expect(got).toStrictEqual(expected);"""
        )

        blocks.append(
            Stripped(
                f"""\
test("over {name}", () => {{
{I}{indent_but_first_line(body_writer.getvalue(), I)}
}});"""
            )
        )

    blocks.append(typescript_common.WARNING)

    writer = io.StringIO()
    for i, block in enumerate(blocks):
        if i > 0:
            writer.write("\n\n")

        writer.write(block)

    writer.write("\n")

    return writer.getvalue()


assert generate.__doc__ is not None
assert generate.__doc__.strip().startswith(__doc__.strip())



"""Generate code to test the JSON de/serialization of concrete classes."""

import io
from typing import List

from icontract import ensure

from aas_core_codegen import intermediate, naming
from aas_core_codegen.common import (
    Stripped,
    Identifier,
)
from aas_core_codegen.golang import common as golang_common, naming as golang_naming
from aas_core_codegen.golang.common import (
    INDENT as I,
    INDENT2 as II,
    INDENT3 as III,
    INDENT4 as IIII,
    INDENT5 as IIIII,
    INDENT6 as IIIIII,
)


def _generate_for_class(cls: intermediate.ConcreteClass) -> List[Stripped]:
    """Generate the tests for the given class."""
    model_type_literal = golang_common.string_literal(naming.json_model_type(cls.name))

    deserialization_function = golang_naming.function_name(
        Identifier(f"{cls.name}_from_jsonable")
    )

    blocks = []  # type: List[Stripped]

    test_name = golang_naming.function_name(
        Identifier(f"Test_{cls.name}_round_trip_OK")
    )

    blocks.append(
        Stripped(
            f"""\
func {test_name}(t *testing.T) {{
{I}pths := aastesting.FindFilesBySuffixRecursively(
{II}filepath.Join(
{III}aastesting.TestDataDir,
{III}"Json",
{III}"Expected",
{III}{model_type_literal},
{II}),
{II}".json",
{I})
{I}sort.Strings(pths)

{I}for _, pth := range pths {{
{II}jsonable := aastesting.MustReadJsonable(
{III}pth,
{II})

{II}deserialized, deseriaErr := aasjsonization.{deserialization_function}(
{III}jsonable,
{II})
{II}ok := assertNoDeserializationError(t, deseriaErr, pth)
{II}if !ok {{
{III}return
{II}}}

{II}anotherJsonable, seriaErr := aasjsonization.ToJsonable(deserialized)
{II}ok = assertNoSerializationError(t, seriaErr, pth)
{II}if !ok {{
{III}return
{II}}}

{II}ok = assertSerializationEqualsDeserialization(
{III}t,
{III}jsonable,
{III}anotherJsonable,
{III}pth,
{II})
{II}if !ok {{
{III}return
{II}}}
{I}}}
}}"""
        )
    )

    test_name = golang_naming.function_name(
        Identifier(f"Test_{cls.name}_deserialization_fail")
    )

    blocks.append(
        Stripped(
            f"""\
func {test_name}(t *testing.T) {{
{I}pattern := filepath.Join(
{II}aastesting.TestDataDir,
{II}"Json",
{II}"Unexpected",
{II}"Unserializable",
{II}"*",  // This asterisk represents the cause.
{II}{model_type_literal},
{I})

{I}causeDirs, err := filepath.Glob(pattern)
{I}if err != nil {{
{II}panic(
{III}fmt.Sprintf(
{IIII}"Failed to find cause directories matching %s: %s",
{IIII}pattern, err.Error(),
{III}),
{II})
{I}}}

{I}for _, causeDir := range causeDirs {{
{II}pths := aastesting.FindFilesBySuffixRecursively(
{III}causeDir,
{III}".json",
{II})
{II}sort.Strings(pths)

{II}for _, pth := range pths {{
{III}jsonable := aastesting.MustReadJsonable(
{IIII}pth,
{III})

{III}relPth, err := filepath.Rel(aastesting.TestDataDir, pth)
{III}if err != nil {{
{IIII}panic(
{IIIII}fmt.Sprintf(
{IIIIII}"Failed to compute the relative path of %s to %s: %s",
{IIIIII}aastesting.TestDataDir, pth, err.Error(),
{IIIII}),
{IIII})
{III}}}

{III}expectedPth := filepath.Join(
{IIII}aastesting.TestDataDir,
{IIII}"DeserializationError",
{IIII}filepath.Dir(relPth),
{IIII}filepath.Base(relPth)+".error",
{III})

{III}_, deseriaErr := aasjsonization.{deserialization_function}(
{IIII}jsonable,
{III})
{III}ok := assertDeserializationErrorEqualsExpectedOrRecord(
{IIII}t, deseriaErr, pth, expectedPth,
{III})
{III}if !ok {{
{IIII}return
{III}}}
{II}}}
{I}}}
}}"""
        )
    )

    return blocks


# fmt: off
@ensure(
    lambda result: result.endswith('\n'),
    "Trailing newline mandatory for valid end-of-files"
)
# fmt: on
def generate(symbol_table: intermediate.SymbolTable, repo_url: Stripped) -> str:
    """Generate code to test the JSON de/serialization of concrete classes."""
    blocks = [
        Stripped("package jsonization_test"),
        golang_common.WARNING,
        Stripped(
            f"""\
import (
{I}"fmt"
{I}"path/filepath"
{I}"sort"
{I}"testing"
{I}aasjsonization "{repo_url}/jsonization"
{I}aastesting "{repo_url}/aastesting"
)"""
        ),
    ]  # type: List[Stripped]

    for concrete_cls in symbol_table.concrete_classes:
        blocks.extend(_generate_for_class(cls=concrete_cls))

    blocks.append(golang_common.WARNING)

    writer = io.StringIO()
    for i, block in enumerate(blocks):
        if i > 0:
            writer.write("\n\n")

        writer.write(block)

    writer.write("\n")

    return writer.getvalue()


assert generate.__doc__ is not None
assert generate.__doc__.strip().startswith(__doc__.strip())



"""Generate the unit tests for Base64-URL functions."""

import io
from typing import List

from icontract import ensure

from aas_core_codegen.common import Stripped
from aas_core_codegen.typescript.common import (
    INDENT as I,
    INDENT2 as II,
)


# fmt: off
@ensure(
    lambda result: result.endswith('\n'),
    "Trailing newline mandatory for valid end-of-files"
)
# fmt: on
def generate() -> str:
    """Generate the unit tests for Base64-URL functions."""
    blocks = [
        Stripped(
            """\
/**
 * Test base64url encoding and decoding.
 */"""
        ),
        Stripped(
            """\
import * as AasCommon from "../src/common";"""
        ),
        Stripped(
            f"""\
function testBase64UrlEncodeDecode(text: string, expectedEncoded: string): void {{
{I}const bytes = Uint8Array.from(text.split("").map((c) => c.charCodeAt(0)));

{I}const encoded = AasCommon.base64UrlEncode(bytes);
{I}expect(encoded).toEqual(expectedEncoded);

{I}const decodedOrError: AasCommon.Either<Uint8Array, string> =
{II}AasCommon.base64UrlDecode(encoded);

{I}expect(decodedOrError.error).toBeNull();
{I}expect(decodedOrError.mustValue()).toEqual(bytes);
}}"""
        ),
        Stripped(
            f"""\
test("empty string", () => {{
{I}testBase64UrlEncodeDecode("", "");
}});"""
        ),
        Stripped(
            f"""\
test("'f' is encoded 'Zg'", () => {{
{I}testBase64UrlEncodeDecode("f", "Zg");
}});"""
        ),
        Stripped(
            f"""\
test("'fo' is encoded 'Zm8'", () => {{
{I}testBase64UrlEncodeDecode("fo", "Zm8");
}});"""
        ),
        Stripped(
            f"""\
test("'foo' is encoded 'Zm9v'", () => {{
{I}testBase64UrlEncodeDecode("foo", "Zm9v");
}});"""
        ),
        Stripped(
            f"""\
test("'foob' is encoded 'Zm9vYg'", () => {{
{I}testBase64UrlEncodeDecode("foob", "Zm9vYg");
}});"""
        ),
        Stripped(
            f"""\
test("'fooba' is encoded 'Zm9vYmE'", () => {{
{I}testBase64UrlEncodeDecode("fooba", "Zm9vYmE");
}});"""
        ),
        Stripped(
            f"""\
test("'foobar' is encoded 'Zm9vYmFy'", () => {{
{I}testBase64UrlEncodeDecode("foobar", "Zm9vYmFy");
}});"""
        ),
        Stripped(
            f"""\
test("RFC 4648 test vectors", () => {{
{I}const testVectors: Array<[string, string]> = [
{II}["", ""],
{II}["f", "Zg"],
{II}["fo", "Zm8"],
{II}["foo", "Zm9v"],
{II}["foob", "Zm9vYg"],
{II}["fooba", "Zm9vYmE"],
{II}["foobar", "Zm9vYmFy"]
{I}];

{I}for (const [input, expected] of testVectors) {{
{II}testBase64UrlEncodeDecode(input, expected);
{I}}}
}});"""
        ),
        Stripped(
            f"""\
test("characters that differ from base64", () => {{
{I}const bytes = new Uint8Array([0x3e, 0x3f, 0xfc, 0xff]);

{I}const base64 = AasCommon.base64Encode(bytes);
{I}const base64url = AasCommon.base64UrlEncode(bytes);

{I}expect(base64).toEqual("Pj/8/w==");
{I}expect(base64url).toEqual("Pj_8_w");
{I}expect(base64url).not.toContain("+");
{I}expect(base64url).not.toContain("/");
{I}expect(base64url).not.toContain("=");
}});"""
        ),
        Stripped(
            f"""\
test("decode with missing padding", () => {{
{I}const testCases = ["Zg", "Zm8", "Zm9vYg", "Zm9vYmE"];

{I}for (const encoded of testCases) {{
{II}const decodedOrError = AasCommon.base64UrlDecode(encoded);
{II}expect(decodedOrError.error).toBeNull();
{I}}}
}});"""
        ),
        Stripped(
            f"""\
test("decode URL-safe characters", () => {{
{I}const encoded = "Pj_8_w";
{I}const decodedOrError = AasCommon.base64UrlDecode(encoded);

{I}expect(decodedOrError.error).toBeNull();
{I}expect(decodedOrError.mustValue()).toEqual(new Uint8Array([0x3e, 0x3f, 0xfc, 0xff]));
}});"""
        ),
        Stripped(
            f"""\
test("round-trip with binary data", () => {{
{I}const bytes = new Uint8Array(256);
{I}for (let i = 0; i < 256; i++) {{
{II}bytes[i] = i;
{I}}}

{I}const encoded = AasCommon.base64UrlEncode(bytes);
{I}const decodedOrError = AasCommon.base64UrlDecode(encoded);

{I}expect(decodedOrError.error).toBeNull();
{I}expect(decodedOrError.mustValue()).toEqual(bytes);
{I}expect(encoded).not.toContain("+");
{I}expect(encoded).not.toContain("/");
{I}expect(encoded).not.toContain("=");
}});"""
        ),
        Stripped(
            f"""\
test("specific byte sequences that produce URL-unsafe characters", () => {{
{I}const testCases: Array<[number[], string]> = [
{II}[[62], "Pg"],
{II}[[63], "Pw"],
{II}[[250], "-g"], // Base64 would start with `+`
{II}[[251], "-w"],
{II}[[252], "_A"], // Base64 would start with `/`
{II}[[253], "_Q"],
{II}[[254], "_g"],
{II}[[255], "_w"],
{II}[[62, 63], "Pj8"],
{II}[[252, 253], "_P0"],
{II}[[254, 255], "_v8"]
{I}];

{I}for (const [numberArray, expectedEncoded] of testCases) {{
{II}const bytes = new Uint8Array(numberArray);
{II}const encoded = AasCommon.base64UrlEncode(bytes);

{II}expect(encoded).toEqual(expectedEncoded);

{II}const decodedOrError = AasCommon.base64UrlDecode(encoded);
{II}expect(decodedOrError.error).toBeNull();
{II}expect(decodedOrError.mustValue()).toEqual(bytes);
{I}}}
}});"""
        ),
        Stripped(
            f"""\
test("invalid characters in input", () => {{
{I}const invalidInputs = ["Zm9v+", "Zm9v/", "Zm9v=", "Zm9v YmFy", "Zm9v\\n", "Zm9v\\t"];

{I}for (const invalid of invalidInputs) {{
{II}const decodedOrError = AasCommon.base64UrlDecode(invalid);
{II}expect(decodedOrError.error).not.toBeNull();
{II}expect(decodedOrError.value).toBeNull();
{I}}}
}});"""
        ),
    ]  # type: List[Stripped]
    writer = io.StringIO()
    for i, block in enumerate(blocks):
        if i > 0:
            writer.write("\n\n")

        writer.write(block)

    writer.write("\n")

    return writer.getvalue()


assert generate.__doc__ is not None
assert generate.__doc__.strip().startswith(__doc__.strip())



"""Generate Java code based on the meta-model."""

import pathlib
from typing import TextIO, Sequence, Tuple, Callable, Optional, List

from aas_core_codegen import specific_implementations, run, java, intermediate
from aas_core_codegen.common import Error
from aas_core_codegen.java import (
    common as java_common,
    lib as java_lib,
    tests as java_tests,
)


def execute(context: run.Context, stdout: TextIO, stderr: TextIO) -> int:
    """Generate the code."""

    verified_ir_table, errors = java_lib.verify_for_types(
        symbol_table=context.symbol_table
    )

    if errors is not None:
        run.write_error_report(
            message=f"Failed to verify the intermediate symbol table "
            f"for generation of Java code"
            f"based on {context.model_path}",
            errors=[context.lineno_columner.error_message(error) for error in errors],
            stderr=stderr,
        )
        return 1

    assert verified_ir_table is not None

    unsupported_contracts_errors = (
        intermediate.errors_if_contracts_for_functions_or_methods_defined(
            verified_ir_table
        )
    )
    if unsupported_contracts_errors is not None:
        run.write_error_report(
            message=f"We do not support pre and post-conditions and snapshots "
            f"at the moment. Please notify the developers if you need this "
            f"feature (based on meta-model {context.model_path})",
            errors=[
                context.lineno_columner.error_message(error)
                for error in unsupported_contracts_errors
            ],
            stderr=stderr,
        )
        return 1

    unsupported_methods_errors = (
        intermediate.errors_if_non_implementation_specific_methods(verified_ir_table)
    )
    if unsupported_methods_errors is not None:
        run.write_error_report(
            message=f"We added some support for understood methods already and keep "
            f"maintaining it as it is only a matter of time when we will "
            f"introduce their transpilation. Introducing them after the fact "
            f"would have been much more difficult.\n"
            f"\n"
            f"At the given moment, however, we deliberately focus only on "
            f"implementation-specific methods. "
            f"(based on meta-model {context.model_path})",
            errors=[
                context.lineno_columner.error_message(error)
                for error in unsupported_methods_errors
            ],
            stderr=stderr,
        )
        return 1

    package_key = specific_implementations.ImplementationKey("package.txt")
    package_text = context.spec_impls.get(package_key, None)
    if package_text is None:
        stderr.write(f"The package snippet is missing: {package_key}\n")
        return 1

    if not java_common.PACKAGE_IDENTIFIER_RE.fullmatch(package_text):
        stderr.write(
            f"The text from the snippet {package_key} "
            f"is not a valid package identifier: {package_text!r}\n"
        )
        return 1

    package = java_common.PackageIdentifier(package_text)

    verify_errors = java_lib.verify_for_verification(
        spec_impls=context.spec_impls,
        verification_functions=verified_ir_table.verification_functions,
    )

    if verify_errors is not None:
        run.write_error_report(
            message="Failed to verify the verification functions for code generation",
            errors=verify_errors,
            stderr=stderr,
        )
        return 1

    project_rel_path = pathlib.Path("src") / "main" / "java" / package.replace(".", "/")
    assert not project_rel_path.is_absolute()

    tests_rel_path = pathlib.Path("src") / "test" / "java"
    assert not tests_rel_path.is_absolute()

    rel_paths_generators: Sequence[
        Tuple[
            pathlib.Path,
            Callable[
                [], Tuple[Optional[List[java_common.JavaFile]], Optional[List[Error]]]
            ],
        ]
    ] = [
        (
            project_rel_path / "constants",
            lambda: java_lib.generate_constants(
                symbol_table=context.symbol_table, package=package
            ),
        ),
        (
            project_rel_path / "copying",
            lambda: java_lib.generate_copying(
                symbol_table=context.symbol_table,
                package=package,
                spec_impls=context.spec_impls,
            ),
        ),
        (
            project_rel_path / "enhancing",
            lambda: java_lib.generate_enhancing(
                symbol_table=context.symbol_table,
                package=package,
                spec_impls=context.spec_impls,
            ),
        ),
        (
            project_rel_path / "generation",
            lambda: java_lib.generate_generation(
                symbol_table=context.symbol_table,
                package=package,
                spec_impls=context.spec_impls,
            ),
        ),
        (
            project_rel_path / "jsonization",
            lambda: java_lib.generate_jsonization(
                symbol_table=context.symbol_table,
                package=package,
                spec_impls=context.spec_impls,
            ),
        ),
        (
            project_rel_path / "reporting",
            lambda: (java_lib.generate_reporting(package=package), None),
        ),
        (
            project_rel_path / "stringification",
            lambda: java_lib.generate_stringification(
                symbol_table=context.symbol_table, package=package
            ),
        ),
        (
            project_rel_path / "types",
            lambda: java_lib.generate_types(
                symbol_table=verified_ir_table,
                package=package,
                spec_impls=context.spec_impls,
            ),
        ),
        (
            project_rel_path / "verification",
            lambda: java_lib.generate_verification(
                symbol_table=verified_ir_table,
                package=package,
                spec_impls=context.spec_impls,
            ),
        ),
        (
            project_rel_path / "visitation",
            lambda: java_lib.generate_visitation(
                symbol_table=context.symbol_table, package=package
            ),
        ),
        (
            project_rel_path / "xmlization",
            lambda: java_lib.generate_xmlization(
                symbol_table=context.symbol_table,
                package=package,
                spec_impls=context.spec_impls,
            ),
        ),
        (
            tests_rel_path,
            lambda: (java_tests.generate_common(package=package), None),
        ),
        (
            tests_rel_path,
            lambda: (java_tests.generate_common_json(package=package), None),
        ),
        (
            tests_rel_path,
            lambda: (
                java_tests.generate_common_jsonization(
                    package=package, symbol_table=context.symbol_table
                ),
                None,
            ),
        ),
        (
            tests_rel_path,
            lambda: (
                java_tests.generate_test_copying(
                    package=package, symbol_table=context.symbol_table
                ),
                None,
            ),
        ),
        (
            tests_rel_path,
            lambda: (
                java_tests.generate_test_descend_and_visitor_through(
                    package=package, symbol_table=context.symbol_table
                ),
                None,
            ),
        ),
        (
            tests_rel_path,
            lambda: (
                java_tests.generate_test_descend_once(
                    package=package, symbol_table=context.symbol_table
                ),
                None,
            ),
        ),
        (
            tests_rel_path,
            lambda: (
                java_tests.generate_test_enhancing(
                    package=package, symbol_table=context.symbol_table
                ),
                None,
            ),
        ),
        (
            tests_rel_path,
            lambda: (
                java_tests.generate_test_jsonization_of_concrete_classes(
                    package=package, symbol_table=context.symbol_table
                ),
                None,
            ),
        ),
        (
            tests_rel_path,
            lambda: (
                java_tests.generate_test_jsonization_of_enums(
                    package=package, symbol_table=context.symbol_table
                ),
                None,
            ),
        ),
        (
            tests_rel_path,
            lambda: (
                java_tests.generate_test_jsonization_of_interfaces(
                    package=package, symbol_table=context.symbol_table
                ),
                None,
            ),
        ),
        (
            tests_rel_path,
            lambda: (
                java_tests.generate_test_over_x_or_empty(
                    package=package, symbol_table=context.symbol_table
                ),
                None,
            ),
        ),
        (
            tests_rel_path,
            lambda: (
                java_tests.generate_test_verification_of_enums(
                    package=package, symbol_table=context.symbol_table
                ),
                None,
            ),
        ),
        (
            tests_rel_path,
            lambda: (
                java_tests.generate_test_x_or_default(
                    package=package, symbol_table=context.symbol_table
                ),
                None,
            ),
        ),
        (
            tests_rel_path,
            lambda: (
                java_tests.generate_test_xmlization_of_concrete_classes(
                    package=package, symbol_table=context.symbol_table
                ),
                None,
            ),
        ),
        (
            tests_rel_path,
            lambda: (
                java_tests.generate_test_xmlization_of_interfaces(
                    package=package, symbol_table=context.symbol_table
                ),
                None,
            ),
        ),
    ]

    for rel_path, generator_func in rel_paths_generators:
        assert not rel_path.is_absolute()

        java_files, errors = generator_func()

        if errors is not None:
            run.write_error_report(
                message=f"Failed to generate {rel_path} "
                f"based on {context.model_path}",
                errors=[
                    context.lineno_columner.error_message(error) for error in errors
                ],
                stderr=stderr,
            )
            return 1

        assert java_files is not None

        for java_file in java_files:
            pth = context.output_dir / rel_path / java_file.name

            try:
                pth.parent.mkdir(parents=True, exist_ok=True)
            except Exception as exception:
                run.write_error_report(
                    message=f"Failed to create the directory {pth.parent}",
                    errors=[str(exception)],
                    stderr=stderr,
                )
                return 1

            try:
                pth.write_text(java_file.content, encoding="utf-8")
            except Exception as exception:
                run.write_error_report(
                    message=f"Failed to write to {pth}",
                    errors=[str(exception)],
                    stderr=stderr,
                )
                return 1

    stdout.write(f"Code generated to: {context.output_dir}\n")
    return 0


assert java.__doc__ == __doc__



"""Generate code to test the ``XOrDefault`` methods."""

from typing import List, Optional

from aas_core_codegen import intermediate, naming
from aas_core_codegen.common import Stripped, indent_but_first_line
from aas_core_codegen.java import common as java_common, naming as java_naming
from aas_core_codegen.java.common import (
    INDENT as I,
    INDENT2 as II,
    INDENT3 as III,
    INDENT4 as IIII,
)


def generate(
    package: java_common.PackageIdentifier,
    symbol_table: intermediate.SymbolTable,
) -> List[java_common.JavaFile]:
    """
    Generate code to test the ``XOrDefault`` methods.
    """
    blocks = [
        Stripped(
            f"""\
private static void compareOrRerecordValue(Object value, Path expectedPath) throws IOException {{
{I}final ObjectMapper objectMapper = new ObjectMapper();
{I}JsonNode got = CommonJson.toJson(value);

{I}if (Common.RECORD_MODE) {{
{II}Files.createDirectories(expectedPath.getParent());
{II}Files.write(expectedPath, got.toString().getBytes());
{I}}}
{I}else
{I}{{
{II}if (!Files.exists(expectedPath)) {{
{III}throw new FileNotFoundException(
{IIII}"The file with the recorded value does not exist: " + expectedPath + "; " +
{IIII}"maybe you want to set the environment variable " +
{IIII}Common.RECORD_MODE_ENVIRONMENT_VARIABLE_NAME + "?");
{II}}}

{II}final JsonNode expected = objectMapper.readTree(expectedPath.toFile());
{II}assertEquals(objectMapper.readTree(expected.toString()), objectMapper.readTree(got.toString()));
{I}}}
}}"""
        )
    ]  # type: List[Stripped]

    for concrete_cls in symbol_table.concrete_classes:
        cls_name_java = java_naming.class_name(concrete_cls.name)
        cls_name_json = naming.json_model_type(concrete_cls.name)

        x_or_default_methods = []  # type: List[intermediate.MethodUnion]
        for method in concrete_cls.methods:
            if method.name.endswith("_or_default"):
                x_or_default_methods.append(method)

        for method in x_or_default_methods:
            method_name_java = java_naming.method_name(method.name)
            method_name_capitalized_java = (
                method_name_java[0].upper() + method_name_java[1:]
            )

            result_enum = None  # type: Optional[intermediate.Enumeration]
            assert method.returns is not None, (
                f"Expected all X_or_default to return something, "
                f"but got None for {concrete_cls.name}.{method.name}"
            )

            if isinstance(
                method.returns, intermediate.OurTypeAnnotation
            ) and isinstance(method.returns.our_type, intermediate.Enumeration):
                result_enum = method.returns.our_type

            if result_enum is None:
                value_assignment_snippet = Stripped(
                    f"final Object value = instance.{method_name_java}();"
                )
            else:
                value_type_name_java = java_naming.enum_name(result_enum.name)
                value_assignment_snippet = Stripped(
                    f"""\
final {value_type_name_java} enumValue = instance.{method_name_java}();
if (enumValue == null) {{
{I}throw new IllegalStateException("Failed to stringify the enum");
}}

final Optional<String> valueOptional = Stringification.toString(enumValue);
assertEquals(true, valueOptional.isPresent());
final String value = valueOptional.get();"""
                )

            # noinspection SpellCheckingInspection
            blocks.append(
                Stripped(
                    f"""\
@Test
public void test{cls_name_java}{method_name_capitalized_java}NonDefault()
{I} throws IOException {{
{I}{cls_name_java} instance = CommonJsonization.loadMaximal{cls_name_java}();

{I}{indent_but_first_line(value_assignment_snippet, I)}

{I}compareOrRerecordValue(
{II}value,
{II}Paths.get(
{III}Common.TEST_DATA_DIR,
{III}"XOrDefault",
{III}{java_common.string_literal(cls_name_json)},
{III}"{method_name_java}.non-default.json"));
}} // public void test{cls_name_java}{method_name_java}NonDefault

@Test
public void test{cls_name_java}{method_name_java}Default()
{I} throws IOException {{
{I}{cls_name_java} instance = CommonJsonization.loadMinimal{cls_name_java}();

{I}{indent_but_first_line(value_assignment_snippet, I)}

{I}compareOrRerecordValue(
{II}value,
{II}Paths.get(
{III}Common.TEST_DATA_DIR,
{III}"XOrDefault",
{III}{java_common.string_literal(cls_name_json)},
{III}"{method_name_java}.default.json"));
}} // public void test{cls_name_java}{method_name_java}Default"""
                )
            )

    blocks_joined = "\n\n".join(blocks)

    return [
        java_common.JavaFile(
            "TestXOrDefault.java",
            f"""\
{java_common.WARNING}

package {package}.tests;

import static org.junit.jupiter.api.Assertions.assertEquals;

import {package}.reporting.*;
import {package}.stringification.Stringification;
import {package}.types.enums.*;
import {package}.types.impl.*;
import com.fasterxml.jackson.databind.JsonNode;
import com.fasterxml.jackson.databind.ObjectMapper;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.Optional;
import org.junit.jupiter.api.Test;

public class TestXOrDefault {{
{I}{indent_but_first_line(blocks_joined, I)}
}} // class TestXOrDefault

// package {package}.tests

{java_common.WARNING}
""",
        )
    ]


assert generate.__doc__ is not None
assert generate.__doc__.strip().startswith(__doc__.strip())



"""Generate code to test ``Is*`` functions."""

import io
from typing import List

from icontract import ensure

from aas_core_codegen import intermediate
from aas_core_codegen.common import (
    Stripped,
    Identifier,
    indent_but_first_line,
)
from aas_core_codegen.golang import common as golang_common, naming as golang_naming
from aas_core_codegen.golang.common import INDENT as I, INDENT2 as II


def _generate_for_cls(
    cls: intermediate.ConcreteClass, symbol_table: intermediate.SymbolTable
) -> Stripped:
    """Generate the test function."""
    must_load_minimal_name = golang_naming.function_name(
        Identifier(f"must_load_minimal_{cls.name}")
    )

    interface_name = golang_naming.interface_name(cls.name)

    block = [
        Stripped(
            f"""\
instance := aastesting.{must_load_minimal_name}()"""
        )
    ]  # type: List[Stripped]

    for concrete_cls in symbol_table.concrete_classes:
        is_function_name = golang_naming.function_name(
            Identifier(f"is_{concrete_cls.name}")
        )

        if cls.is_subclass_of(concrete_cls):
            block.append(
                Stripped(
                    f"""\
if !aastypes.{is_function_name}(instance) {{
{I}t.Errorf(
{II}"Expected {is_function_name} to be true on an instance " +
{II}"of {interface_name} with runtime type %T and with model type %v",
{II}instance, instance.ModelType(),
{I})
}}"""
                )
            )
        else:
            block.append(
                Stripped(
                    f"""\
if aastypes.{is_function_name}(instance) {{
{I}t.Errorf(
{II}"Expected {is_function_name} to be false on an instance " +
{II}"of {interface_name} with runtime type %T and with model type %v",
{II}instance, instance.ModelType(),
{I})
}}"""
                )
            )

    body = "\n\n".join(block)

    test_name = golang_naming.function_name(
        Identifier(f"test_is_Xxx_on_an_instance_of_{cls.name}")
    )

    return Stripped(
        f"""\
func {test_name}(t *testing.T) {{
{I}{indent_but_first_line(body, I)}
}}"""
    )


# fmt: off
@ensure(
    lambda result: result.endswith('\n'),
    "Trailing newline mandatory for valid end-of-files"
)
# fmt: on
def generate(symbol_table: intermediate.SymbolTable, repo_url: Stripped) -> str:
    """Generate code to test ``Is*`` functions."""
    blocks = [
        Stripped("package types_is_xxx_test"),
        golang_common.WARNING,
        Stripped(
            f"""\
import (
{I}"testing"
{I}aastesting "{repo_url}/aastesting"
{I}aastypes "{repo_url}/types"
)"""
        ),
    ]  # type: List[Stripped]

    for concrete_cls in symbol_table.concrete_classes:
        blocks.append(_generate_for_cls(cls=concrete_cls, symbol_table=symbol_table))

    blocks.append(golang_common.WARNING)

    writer = io.StringIO()
    for i, block in enumerate(blocks):
        if i > 0:
            writer.write("\n\n")

        writer.write(block)

    writer.write("\n")

    return writer.getvalue()


assert generate.__doc__ is not None
assert generate.__doc__.strip().startswith(__doc__.strip())



"""Generate the unit test code for C++."""

from aas_core_codegen.cpp.tests import (
    _generate_common,
    _generate_common_examples,
    _generate_common_jsonization,
    _generate_common_xmlization,
    _generate_test_descent_and_descent_once,
    _generate_test_jsonization_dispatch,
    _generate_test_jsonization_of_concrete_classes,
    _generate_test_revm,
    _generate_test_stringification_base64,
    _generate_test_stringification_of_enums,
    _generate_test_verification,
    _generate_test_wstringification_of_enums,
    _generate_test_xmlization_dispatch,
    _generate_test_xmlization_of_concrete_classes,
    _generate_test_x_or_default,
)

generate_common_header = _generate_common.generate_header
generate_common_implementation = _generate_common.generate_implementation

generate_common_examples_header = _generate_common_examples.generate_header
generate_common_examples_implementation = (
    _generate_common_examples.generate_implementation
)

generate_common_jsonization_header = _generate_common_jsonization.generate_header
generate_common_jsonization_implementation = (
    _generate_common_jsonization.generate_implementation
)

generate_common_xmlization_header = _generate_common_xmlization.generate_header
generate_common_xmlization_implementation = (
    _generate_common_xmlization.generate_implementation
)

generate_test_descent_and_descent_once_implementation = (
    _generate_test_descent_and_descent_once.generate_implementation
)
generate_test_jsonization_dispatch_implementation = (
    _generate_test_jsonization_dispatch.generate_implementation
)
generate_test_jsonization_of_concrete_classes_implementation = (
    _generate_test_jsonization_of_concrete_classes.generate_implementation
)
generate_test_revm_implementation = _generate_test_revm.generate_implementation
generate_test_stringification_base64_implementation = (
    _generate_test_stringification_base64.generate_implementation
)
generate_test_stringification_of_enums_implementation = (
    _generate_test_stringification_of_enums.generate_implementation
)
generate_test_verification_implementation = (
    _generate_test_verification.generate_implementation
)
generate_test_wstringification_of_enums_implementation = (
    _generate_test_wstringification_of_enums.generate_implementation
)
generate_test_xmlization_dispatch_implementation = (
    _generate_test_xmlization_dispatch.generate_implementation
)
generate_test_xmlization_of_concrete_classes_implementation = (
    _generate_test_xmlization_of_concrete_classes.generate_implementation
)
generate_test_x_or_default_implementation = (
    _generate_test_x_or_default.generate_implementation
)



"""Generate TypeScript code based on the meta-model."""

import pathlib
from typing import TextIO, Sequence, Tuple, Callable, Optional, List

from aas_core_codegen import run, intermediate, typescript
from aas_core_codegen.common import Error
from aas_core_codegen.typescript import lib as typescript_lib, tests as typescript_tests


def execute(context: run.Context, stdout: TextIO, stderr: TextIO) -> int:
    """Generate code."""
    verified_ir_table, errors = typescript_lib.verify_for_types(
        symbol_table=context.symbol_table
    )

    if errors is not None:
        run.write_error_report(
            message=f"Failed to verify the intermediate symbol table "
            f"for generation of TypeScript code"
            f"based on {context.model_path}",
            errors=[context.lineno_columner.error_message(error) for error in errors],
            stderr=stderr,
        )
        return 1

    assert verified_ir_table is not None

    unsupported_contracts_errors = (
        intermediate.errors_if_contracts_for_functions_or_methods_defined(
            verified_ir_table
        )
    )
    if unsupported_contracts_errors is not None:
        run.write_error_report(
            message=f"We do not support pre and post-conditions and snapshots "
            f"at the moment. Please notify the developers if you need this "
            f"feature (based on meta-model {context.model_path})",
            errors=[
                context.lineno_columner.error_message(error)
                for error in unsupported_contracts_errors
            ],
            stderr=stderr,
        )
        return 1

    unsupported_methods_errors = (
        intermediate.errors_if_non_implementation_specific_methods(verified_ir_table)
    )
    if unsupported_methods_errors is not None:
        run.write_error_report(
            message=f"We added some support for understood methods already and keep "
            f"maintaining it as it is only a matter of time when we will "
            f"introduce their transpilation. Introducing them after the fact "
            f"would have been much more difficult.\n"
            f"\n"
            f"At the given moment, however, we deliberately focus only on "
            f"implementation-specific methods. "
            f"(based on meta-model {context.model_path})",
            errors=[
                context.lineno_columner.error_message(error)
                for error in unsupported_methods_errors
            ],
            stderr=stderr,
        )
        return 1

    verification_functions_errors = typescript_lib.verify_verification_functions(
        spec_impls=context.spec_impls,
        verification_functions=verified_ir_table.verification_functions,
    )

    if verification_functions_errors is not None:
        run.write_error_report(
            message=(
                "Failed to verify for the generation of TypeScript verification "
                "functions code"
            ),
            errors=verification_functions_errors,
            stderr=stderr,
        )
        return 1

    src_rel_path = pathlib.Path("src")
    assert not src_rel_path.is_absolute()

    test_rel_path = pathlib.Path("test")
    assert not test_rel_path.is_absolute()

    rel_paths_generators: Sequence[
        Tuple[pathlib.Path, Callable[[], Tuple[Optional[str], Optional[List[Error]]]]]
    ] = [
        (src_rel_path / "common.ts", lambda: (typescript_lib.generate_common(), None)),
        (
            src_rel_path / "constants.ts",
            lambda: typescript_lib.generate_constants(
                symbol_table=context.symbol_table
            ),
        ),
        (
            src_rel_path / "index.ts",
            lambda: typescript_lib.generate_index(
                spec_impls=context.spec_impls,
            ),
        ),
        (
            src_rel_path / "jsonization.ts",
            lambda: typescript_lib.generate_jsonization(
                symbol_table=context.symbol_table,
                spec_impls=context.spec_impls,
            ),
        ),
        (
            src_rel_path / "xmlization.ts",
            lambda: typescript_lib.generate_xmlization(
                symbol_table=context.symbol_table,
                spec_impls=context.spec_impls,
            ),
        ),
        (
            src_rel_path / "stringification.ts",
            lambda: typescript_lib.generate_stringification(
                symbol_table=context.symbol_table
            ),
        ),
        (
            src_rel_path / "types.ts",
            lambda: typescript_lib.generate_types(
                symbol_table=verified_ir_table,
                spec_impls=context.spec_impls,
            ),
        ),
        (
            src_rel_path / "verification.ts",
            lambda: typescript_lib.generate_verification(
                symbol_table=verified_ir_table,
                spec_impls=context.spec_impls,
            ),
        ),
        (
            test_rel_path / "common.ts",
            lambda: typescript_tests.generate_common(spec_impls=context.spec_impls),
        ),
        (
            test_rel_path / "common.base64.spec.ts",
            lambda: (
                typescript_tests.generate_common_base64_spec(),
                None,
            ),
        ),
        (
            test_rel_path / "common.base64url.spec.ts",
            lambda: (
                typescript_tests.generate_common_base64url_spec(),
                None,
            ),
        ),
        (
            test_rel_path / "commonJsonization.ts",
            lambda: (
                typescript_tests.generate_common_jsonization(
                    symbol_table=verified_ir_table,
                ),
                None,
            ),
        ),
        (
            test_rel_path / "commonXmlization.ts",
            lambda: (
                typescript_tests.generate_common_xmlization(
                    symbol_table=verified_ir_table,
                ),
                None,
            ),
        ),
        (
            test_rel_path / "jsonization.concreteClasses.spec.ts",
            lambda: (
                typescript_tests.generate_jsonization_concrete_classes_spec(
                    symbol_table=verified_ir_table,
                ),
                None,
            ),
        ),
        (
            test_rel_path / "jsonization.enums.spec.ts",
            lambda: (
                typescript_tests.generate_jsonization_enums_spec(
                    symbol_table=verified_ir_table,
                ),
                None,
            ),
        ),
        (
            test_rel_path / "jsonization.interfaces.spec.ts",
            lambda: (
                typescript_tests.generate_jsonization_interfaces_spec(
                    symbol_table=verified_ir_table,
                ),
                None,
            ),
        ),
        (
            test_rel_path / "xmlization.concreteClasses.spec.ts",
            lambda: (
                typescript_tests.generate_xmlization_concrete_classes_spec(
                    symbol_table=verified_ir_table,
                ),
                None,
            ),
        ),
        (
            test_rel_path / "xmlization.enums.spec.ts",
            lambda: (
                typescript_tests.generate_xmlization_enums_spec(
                    symbol_table=verified_ir_table,
                ),
                None,
            ),
        ),
        (
            test_rel_path / "xmlization.interfaces.spec.ts",
            lambda: (
                typescript_tests.generate_xmlization_interfaces_spec(
                    symbol_table=verified_ir_table,
                ),
                None,
            ),
        ),
        (
            test_rel_path / "types.casts.spec.ts",
            lambda: (
                typescript_tests.generate_types_casts_spec(
                    symbol_table=verified_ir_table,
                ),
                None,
            ),
        ),
        (
            test_rel_path / "types.descendAndPassThroughVisitor.spec.ts",
            lambda: (
                typescript_tests.generate_types_descend_and_pass_through_visitor_spec(
                    symbol_table=verified_ir_table,
                ),
                None,
            ),
        ),
        (
            test_rel_path / "types.descendOnce.spec.ts",
            lambda: (
                typescript_tests.generate_types_descend_once_spec(
                    symbol_table=verified_ir_table,
                ),
                None,
            ),
        ),
        (
            test_rel_path / "types.modelType.spec.ts",
            lambda: (
                typescript_tests.generate_types_model_type_spec(
                    symbol_table=verified_ir_table,
                ),
                None,
            ),
        ),
        (
            test_rel_path / "types.overEnum.spec.ts",
            lambda: (
                typescript_tests.generate_types_over_enum_spec(
                    symbol_table=verified_ir_table,
                ),
                None,
            ),
        ),
        (
            test_rel_path / "types.overXOrEmpty.spec.ts",
            lambda: (
                typescript_tests.generate_types_over_x_or_empty_spec(
                    symbol_table=verified_ir_table,
                ),
                None,
            ),
        ),
        (
            test_rel_path / "types.typeMatches.spec.ts",
            lambda: (
                typescript_tests.generate_types_type_matches_spec(
                    symbol_table=verified_ir_table,
                ),
                None,
            ),
        ),
        (
            test_rel_path / "types.xOrDefault.spec.ts",
            lambda: (
                typescript_tests.generate_types_x_or_default_spec(
                    symbol_table=verified_ir_table,
                ),
                None,
            ),
        ),
    ]

    for rel_path, generator_func in rel_paths_generators:
        assert not rel_path.is_absolute()

        code, errors = generator_func()

        if errors is not None:
            run.write_error_report(
                message=f"Failed to generate {rel_path} "
                f"based on {context.model_path}",
                errors=[
                    context.lineno_columner.error_message(error) for error in errors
                ],
                stderr=stderr,
            )
            return 1

        assert code is not None

        pth = context.output_dir / rel_path

        try:
            pth.parent.mkdir(parents=True, exist_ok=True)
        except Exception as exception:
            run.write_error_report(
                message=f"Failed to create the directory {pth.parent}",
                errors=[str(exception)],
                stderr=stderr,
            )
            return 1

        try:
            pth.write_text(code, encoding="utf-8")
        except Exception as exception:
            run.write_error_report(
                message=f"Failed to write to {pth}",
                errors=[str(exception)],
                stderr=stderr,
            )
            return 1

    stdout.write(f"Code generated to: {context.output_dir}\n")
    return 0


assert typescript.__doc__ == __doc__



"""Generate code for common XML de/serialization shared across the tests."""

import io
from typing import List

from icontract import ensure

from aas_core_codegen import intermediate, naming
from aas_core_codegen.common import Stripped, Identifier
from aas_core_codegen.typescript import (
    common as typescript_common,
    naming as typescript_naming,
)
from aas_core_codegen.typescript.common import (
    INDENT as I,
    INDENT2 as II,
    INDENT3 as III,
)


# fmt: off
@ensure(
    lambda result: result.endswith('\n'),
    "Trailing newline mandatory for valid end-of-files"
)
# fmt: on
def generate(symbol_table: intermediate.SymbolTable) -> str:
    """Generate code for common XML de/serialization shared across the tests."""
    blocks = [
        Stripped(
            """\
/**
 * Provide functions for loading generated XML examples of AAS instances.
 */"""
        ),
        typescript_common.WARNING,
        Stripped(
            """\
import * as fs from "fs";
import * as path from "path";

import * as AasTypes from "../src/types";
import * as AasXmlization from "../src/xmlization";

import * as TestCommon from "./common";"""
        ),
    ]  # type: List[Stripped]

    for concrete_cls in symbol_table.concrete_classes:
        cls_name_typescript = typescript_naming.class_name(concrete_cls.name)
        cls_name_xml = naming.xml_class_name(concrete_cls.name)

        load_maximal_name = typescript_naming.function_name(
            Identifier(f"load_maximal_{concrete_cls.name}")
        )

        load_minimal_name = typescript_naming.function_name(
            Identifier(f"load_minimal_{concrete_cls.name}")
        )

        as_function = typescript_naming.function_name(
            Identifier(f"as_{concrete_cls.name}")
        )

        blocks.append(
            Stripped(
                f"""\
/**
 * Load a maximal XML example of {{@link types.{cls_name_typescript}}}
 * from the test data directory.
 */
export function {load_maximal_name}(
): AasTypes.{cls_name_typescript} {{
{I}const aPath = path.join(
{II}TestCommon.TEST_DATA_DIR,
{II}"Xml",
{II}"Expected",
{II}{typescript_common.string_literal(cls_name_xml)},
{II}"maximal.xml"
{I});

{I}const text = fs.readFileSync(aPath, "utf-8");

{I}const instanceOrError = AasXmlization.fromXmlString(
{II}text
{I});
{I}expect(instanceOrError.error).toBeNull();
{I}const instance = instanceOrError.mustValue();

{I}const casted = AasTypes.{as_function}(instance);
{I}if (casted === null) {{
{II}throw new Error(
{III}`Expected instance of {cls_name_typescript} in ${{aPath}}, ` +
{III}`but got: ${{typeof instance}}`
{II});
{I}}}
{I}return casted;
}}"""
            )
        )

        blocks.append(
            Stripped(
                f"""\
/**
 * Load a minimal XML example of {{@link types.{cls_name_typescript}}}
 * from the test data directory.
 */
export function {load_minimal_name}(
): AasTypes.{cls_name_typescript} {{
{I}const aPath = path.join(
{II}TestCommon.TEST_DATA_DIR,
{II}"Xml",
{II}"Expected",
{II}{typescript_common.string_literal(cls_name_xml)},
{II}"minimal.xml"
{I});

{I}const text = fs.readFileSync(aPath, "utf-8");

{I}const instanceOrError = AasXmlization.fromXmlString(
{II}text
{I});
{I}expect(instanceOrError.error).toBeNull();
{I}const instance = instanceOrError.mustValue();

{I}const casted = AasTypes.{as_function}(instance);
{I}if (casted === null) {{
{II}throw new Error(
{III}`Expected instance of {cls_name_typescript} in ${{aPath}}, ` +
{III}`but got: ${{typeof instance}}`
{II});
{I}}}
{I}return casted;
}}"""
            )
        )

    blocks.append(typescript_common.WARNING)

    writer = io.StringIO()
    for i, block in enumerate(blocks):
        if i > 0:
            writer.write("\n\n")

        writer.write(block)

    writer.write("\n")

    return writer.getvalue()


assert generate.__doc__ is not None
assert generate.__doc__.strip().startswith(__doc__.strip())



"""Generate code to test the JSON de/serialization of enumerations."""

import json
from typing import List

from aas_core_codegen import intermediate, naming
from aas_core_codegen.common import Stripped, indent_but_first_line
from aas_core_codegen.java import common as java_common, naming as java_naming
from aas_core_codegen.java.common import (
    INDENT as I,
    INDENT2 as II,
)


def generate(
    package: java_common.PackageIdentifier,
    symbol_table: intermediate.SymbolTable,
) -> List[java_common.JavaFile]:
    """
    Generate code to test the JSON de/serialization of enumerations.
    """
    blocks = []  # type: List[Stripped]

    for enumeration in symbol_table.enumerations:
        enum_name = java_naming.enum_name(enumeration.name)
        enum_name_camel_case = naming.lower_camel_case(enumeration.name)

        assert (
            len(enumeration.literals) > 0
        ), f"Unexpected enumeration without literals: {enumeration.name}"

        literal_value = enumeration.literals[0].value
        literal_value_json_str = json.dumps(literal_value)

        blocks.append(
            Stripped(
                f"""\
@Test
public void testRoundTrip{enum_name}() {{
{I}final JsonNode node = JsonNodeFactory.instance.textNode(
{II}{java_common.string_literal(literal_value)});

{I}final {enum_name} parsed = Jsonization.Deserialize.deserialize{enum_name}(node);

{I}final JsonNode serialized = Jsonization.Serialize.{enum_name_camel_case}ToJsonValue(parsed);

{I}assertEquals(
{II}{java_common.string_literal(literal_value_json_str)},
{II}serialized.toString());
}} // void testRoundTrip{enum_name}"""
            )
        )

    blocks_joined = "\n\n".join(blocks)

    return [
        java_common.JavaFile(
            "TestJsonizationOfEnums.java",
            f"""\
{java_common.WARNING}

package {package}.tests;

import static org.junit.jupiter.api.Assertions.assertEquals;

import {package}.jsonization.Jsonization;
import {package}.types.enums.*;
import com.fasterxml.jackson.databind.JsonNode;
import com.fasterxml.jackson.databind.node.JsonNodeFactory;
import org.junit.jupiter.api.Test;

public class TestJsonizationOfEnums {{
{I}{indent_but_first_line(blocks_joined, I)}
}} // class TestJsonizationOfEnums

// package {package}.tests

{java_common.WARNING}
""",
        )
    ]


assert generate.__doc__ is not None
assert generate.__doc__.strip().startswith(__doc__.strip())



"""Generate Golang code based on the meta-model."""

import pathlib
from typing import TextIO, Sequence, Tuple, Callable, Optional, List

from aas_core_codegen import run, intermediate, specific_implementations, golang
from aas_core_codegen.common import Stripped, Error
from aas_core_codegen.golang import lib as golang_lib, tests as golang_tests


def execute(context: run.Context, stdout: TextIO, stderr: TextIO) -> int:
    """Generate code."""
    verified_ir_table, errors = golang_lib.verify_for_types(
        symbol_table=context.symbol_table
    )

    if errors is not None:
        run.write_error_report(
            message=f"Failed to verify the intermediate symbol table "
            f"for generation of Golang code"
            f"based on {context.model_path}",
            errors=[context.lineno_columner.error_message(error) for error in errors],
            stderr=stderr,
        )
        return 1

    assert verified_ir_table is not None

    unsupported_contracts_errors = (
        intermediate.errors_if_contracts_for_functions_or_methods_defined(
            verified_ir_table
        )
    )
    if unsupported_contracts_errors is not None:
        run.write_error_report(
            message=f"We do not support pre and post-conditions and snapshots "
            f"at the moment. Please notify the developers if you need this "
            f"feature (based on meta-model {context.model_path})",
            errors=[
                context.lineno_columner.error_message(error)
                for error in unsupported_contracts_errors
            ],
            stderr=stderr,
        )
        return 1

    unsupported_methods_errors = (
        intermediate.errors_if_non_implementation_specific_methods(verified_ir_table)
    )
    if unsupported_methods_errors is not None:
        run.write_error_report(
            message=f"We added some support for understood methods already and keep "
            f"maintaining it as it is only a matter of time when we will "
            f"introduce their transpilation. Introducing them after the fact "
            f"would have been much more difficult.\n"
            f"\n"
            f"At the given moment, however, we deliberately focus only on "
            f"implementation-specific methods. "
            f"(based on meta-model {context.model_path})",
            errors=[
                context.lineno_columner.error_message(error)
                for error in unsupported_methods_errors
            ],
            stderr=stderr,
        )
        return 1

    verification_functions_errors = golang_lib.verify_verification_functions(
        spec_impls=context.spec_impls,
        verification_functions=verified_ir_table.verification_functions,
    )

    if verification_functions_errors is not None:
        run.write_error_report(
            message="Failed to verify for the generation of Golang verification code",
            errors=verification_functions_errors,
            stderr=stderr,
        )
        return 1

    # region Repo URL

    repo_url_key = specific_implementations.ImplementationKey("repo_url.txt")
    repo_url_text = context.spec_impls.get(repo_url_key, None)
    if repo_url_text is None:
        stderr.write(f"The repo URL snippet is missing: {repo_url_key}\n")
        return 1

    repo_url = Stripped(repo_url_text.strip())

    # endregion

    base_rel_path = pathlib.Path(".")
    assert not base_rel_path.is_absolute()

    rel_paths_generators: Sequence[
        Tuple[pathlib.Path, Callable[[], Tuple[Optional[str], Optional[List[Error]]]]]
    ] = [
        (
            base_rel_path / "common/common.go",
            lambda: (golang_lib.generate_common(), None),
        ),
        (
            base_rel_path / "constants/constants.go",
            lambda: golang_lib.generate_constants(
                symbol_table=context.symbol_table, repo_url=repo_url
            ),
        ),
        (
            base_rel_path / "enhancing/enhancing.go",
            lambda: golang_lib.generate_enhancing(
                symbol_table=context.symbol_table,
                spec_impls=context.spec_impls,
                repo_url=repo_url,
            ),
        ),
        (
            base_rel_path / "jsonization/jsonization.go",
            lambda: golang_lib.generate_jsonization(
                symbol_table=context.symbol_table,
                spec_impls=context.spec_impls,
                repo_url=repo_url,
            ),
        ),
        (
            base_rel_path / "reporting/reporting.go",
            lambda: (golang_lib.generate_reporting(), None),
        ),
        (
            base_rel_path / "stringification/stringification.go",
            lambda: golang_lib.generate_stringification(
                symbol_table=context.symbol_table, repo_url=repo_url
            ),
        ),
        (
            base_rel_path / "types/types.go",
            lambda: golang_lib.generate_types(
                symbol_table=verified_ir_table, spec_impls=context.spec_impls
            ),
        ),
        (
            base_rel_path / "verification/verification.go",
            lambda: golang_lib.generate_verification(
                symbol_table=verified_ir_table,
                spec_impls=context.spec_impls,
                repo_url=repo_url,
            ),
        ),
        (
            base_rel_path / "xmlization/xmlization.go",
            lambda: golang_lib.generate_xmlization(
                symbol_table=context.symbol_table,
                spec_impls=context.spec_impls,
                repo_url=repo_url,
            ),
        ),
        (
            base_rel_path / "aastesting/common_jsonization.go",
            lambda: (
                golang_tests.generate_aastesting_common_jsonization(
                    symbol_table=verified_ir_table, repo_url=repo_url
                ),
                None,
            ),
        ),
        (
            base_rel_path / "aastesting/constants.go",
            lambda: (
                golang_tests.generate_aastesting_constants(repo_url=repo_url),
                None,
            ),
        ),
        (
            base_rel_path / "aastesting/deep_equal.go",
            lambda: (
                golang_tests.generate_aastesting_deep_equal(
                    symbol_table=verified_ir_table, repo_url=repo_url
                ),
                None,
            ),
        ),
        (
            base_rel_path / "aastesting/doc.go",
            lambda: (golang_tests.generate_aastesting_doc(), None),
        ),
        (
            base_rel_path / "aastesting/filesystem.go",
            lambda: (golang_tests.generate_aastesting_filesystem(), None),
        ),
        (
            base_rel_path / "aastesting/tracing.go",
            lambda: (
                golang_tests.generate_aastesting_tracing(repo_url=repo_url),
                None,
            ),
        ),
        (
            base_rel_path / "enhancing/test/enhancing_test.go",
            lambda: (
                golang_tests.generate_enhancing_test(
                    symbol_table=verified_ir_table, repo_url=repo_url
                ),
                None,
            ),
        ),
        (
            base_rel_path / "jsonization/test/classes_with_descendants_test.go",
            lambda: (
                golang_tests.generate_jsonization_test_classes_with_descendants_test(
                    symbol_table=verified_ir_table, repo_url=repo_url
                ),
                None,
            ),
        ),
        (
            base_rel_path / "jsonization/test/common_test.go",
            lambda: (
                golang_tests.generate_jsonization_test_common_test(repo_url=repo_url),
                None,
            ),
        ),
        (
            base_rel_path / "jsonization/test/concrete_classes_test.go",
            lambda: (
                golang_tests.generate_jsonization_test_concrete_classes_test(
                    symbol_table=verified_ir_table, repo_url=repo_url
                ),
                None,
            ),
        ),
        (
            base_rel_path / "jsonization/test/enums_test.go",
            lambda: (
                golang_tests.generate_jsonization_test_enums_test(
                    symbol_table=verified_ir_table, repo_url=repo_url
                ),
                None,
            ),
        ),
        (
            base_rel_path / "types/descend_test/common.go",
            lambda: (
                golang_tests.generate_descend_test_common(repo_url=repo_url),
                None,
            ),
        ),
        (
            base_rel_path / "types/descend_test/descend_once_test.go",
            lambda: (
                golang_tests.generate_descend_test_descend_once_test(
                    symbol_table=verified_ir_table, repo_url=repo_url
                ),
                None,
            ),
        ),
        (
            base_rel_path / "types/descend_test/descend_test.go",
            lambda: (
                golang_tests.generate_descend_test_descend_test(
                    symbol_table=verified_ir_table, repo_url=repo_url
                ),
                None,
            ),
        ),
        (
            base_rel_path / "types/is_xxx_test/is_xxx_test.go",
            lambda: (
                golang_tests.generate_is_xxx_test(
                    symbol_table=verified_ir_table, repo_url=repo_url
                ),
                None,
            ),
        ),
        (
            base_rel_path / "types/xxx_or_default_test/xxx_or_default_test.go",
            lambda: (
                golang_tests.generate_xxx_or_default_test(
                    symbol_table=verified_ir_table, repo_url=repo_url
                ),
                None,
            ),
        ),
        (
            base_rel_path / "verification/test/verification_test.go",
            lambda: (
                golang_tests.generate_verification_test(
                    symbol_table=verified_ir_table, repo_url=repo_url
                ),
                None,
            ),
        ),
        (
            base_rel_path / "xmlization/test/common_test.go",
            lambda: (
                golang_tests.generate_xmlization_test_common_test(repo_url=repo_url),
                None,
            ),
        ),
        (
            base_rel_path / "xmlization/test/concrete_classes_test.go",
            lambda: (
                golang_tests.generate_xmlization_test_concrete_classes_test(
                    symbol_table=verified_ir_table, repo_url=repo_url
                ),
                None,
            ),
        ),
    ]

    for rel_path, generator_func in rel_paths_generators:
        assert not rel_path.is_absolute()

        code, errors = generator_func()

        if errors is not None:
            run.write_error_report(
                message=f"Failed to generate {rel_path} "
                f"based on {context.model_path}",
                errors=[
                    context.lineno_columner.error_message(error) for error in errors
                ],
                stderr=stderr,
            )
            return 1

        assert code is not None

        pth = context.output_dir / rel_path

        try:
            pth.parent.mkdir(parents=True, exist_ok=True)
        except Exception as exception:
            run.write_error_report(
                message=f"Failed to create the directory {pth.parent}",
                errors=[str(exception)],
                stderr=stderr,
            )
            return 1

        try:
            pth.write_text(code, encoding="utf-8")
        except Exception as exception:
            run.write_error_report(
                message=f"Failed to write to {pth}",
                errors=[str(exception)],
                stderr=stderr,
            )
            return 1

    stdout.write(f"Code generated to: {context.output_dir}\n")
    return 0


assert golang.__doc__ == __doc__
